# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ..command import RobotoCommand
from ..context import CLIContext


def where(args, context: CLIContext, parser: argparse.ArgumentParser):
    cache_dir = context.roboto_config.get_cache_dir()
    print(cache_dir)


def where_setup_parser(parser):
    pass


where_command = RobotoCommand(
    name="where",
    logic=where,
    setup_parser=where_setup_parser,
    command_kwargs={"help": "Prints the path to the Roboto cache directory."},
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Action runtime setup utilities.

This module provides utilities for preparing action runtime (invocation) environments,
both in hosted execution and local development contexts.
"""

import collections.abc
import json
import pathlib
import typing

from ..domain import actions, files, secrets
from ..exceptions import (
    RobotoDomainException,
    RobotoNotFoundException,
    RobotoUnauthorizedException,
)
from ..http import RobotoClient
from ..logging import default_logger
from ..roboto_search import RobotoSearch
from .action_input import (
    ActionInputRecord,
    ActionInputResolver,
)
from .exceptions import PrepareEnvException
from .exit_codes import ExitCode

log = default_logger()


def _ensure_file_path_exists(p: pathlib.Path):
    if p.exists():
        return

    if not p.parent.exists():
        p.parent.mkdir(parents=True)

    p.touch()


def _merge_parameters_with_defaults(
    action_parameters: collections.abc.Iterable[actions.ActionParameter],
    parameter_values: typing.Optional[collections.abc.Mapping[str, typing.Any]],
) -> dict[str, typing.Any]:
    """Merge supplied parameter values with action parameter defaults."""
    log.info("Merging action parameters with defaults")
    supplied_values = parameter_values or {}

    # Start with defaults for all parameters that have them
    merged = {param.name: param.default for param in action_parameters if param.default is not None}

    # Override with supplied values
    merged |= supplied_values

    return merged


def prepare_invocation_parameters(
    action_parameters: collections.abc.Iterable[actions.ActionParameter],
    provided_parameter_values: collections.abc.Mapping[str, typing.Any],
    parameters_values_file: pathlib.Path,
    secrets_file: pathlib.Path,
    org_id: typing.Optional[str],
    roboto_client: RobotoClient,
):
    merged_parameters = _merge_parameters_with_defaults(action_parameters, provided_parameter_values)

    secret_params: dict[str, str] = {}
    for param_name, param_value in merged_parameters.items():
        if secrets.is_secret_uri(param_value):
            log.info(
                "Param '%s' is a secret, resolving value",
                param_name,
            )
            try:
                secret = secrets.Secret.from_uri(
                    uri=param_value,
                    roboto_client=roboto_client,
                    fallback_org_id=org_id,
                )
                actual_value = secret.read_value().get_secret_value()
                secret_params[param_name] = actual_value
            except (RobotoNotFoundException, RobotoUnauthorizedException):
                raise PrepareEnvException(
                    ExitCode.InternalError,
                    f"Failed to retrieve secret {param_value}, because it does not exist.",
                ) from None

    if secret_params:
        log.info("Writing secret action parameter values to %s", secrets_file)
        _ensure_file_path_exists(secrets_file)
        secrets_file.write_text(json.dumps(secret_params, indent=2))

    log.info(
        "Writing action parameter values to %s",
        parameters_values_file,
    )
    _ensure_file_path_exists(parameters_values_file)
    parameters_values_file.write_text(json.dumps(merged_parameters, indent=2))


def _hardlink_resolved_input_to_target_directory(
    resolved_input: ActionInputRecord,
    target_directory: pathlib.Path,
) -> ActionInputRecord:
    files_relative_to_target_dir: list[tuple[files.FileRecord, typing.Optional[pathlib.Path]]] = []

    unique_dataset_count = len(set([record.association_id for record, _ in resolved_input.files]))
    for record, cached_path in resolved_input.files:
        if not cached_path:
            files_relative_to_target_dir.append((record, None))
            continue

        # GM (2025-11-26):
        # Fix for ENG-1848: Handle multi-dataset input files
        # Previously: files downloaded to `<input_dir>/<relative_path>`
        # Now: files from multiple datasets prefixed with `<dataset_id>/<relative_path>`
        # Single dataset maintains backward compatibility to avoid breaking existing actions
        # that do not use `InvocationContext.get_input()`
        if unique_dataset_count > 1:
            target = target_directory / record.association_id / record.relative_path
        else:
            # backward compat behavior
            target = target_directory / record.relative_path

        target.parent.mkdir(parents=True, exist_ok=True)
        target.hardlink_to(cached_path)

        files_relative_to_target_dir.append((record, target))

    return resolved_input.model_copy(update={"files": files_relative_to_target_dir})


def prepare_invocation_input_data(
    requires_downloaded_inputs: bool,
    input_data: typing.Optional[actions.InvocationInput],
    target_directory: pathlib.Path,
    download_directory: pathlib.Path,
    inputs_data_manifest_file: pathlib.Path,
    roboto_client: RobotoClient,
    roboto_search: RobotoSearch,
):
    # Resolve and optionally download inputs
    resolved_input: typing.Optional[ActionInputRecord] = None
    if input_data is not None:
        log.info("Resolving input data")

        input_resolver = ActionInputResolver.from_env(
            roboto_client=roboto_client,
            roboto_search=roboto_search,
        )

        try:
            resolved_input = input_resolver.resolve_input_spec(
                input_data,
                download=requires_downloaded_inputs,
                download_path=download_directory,
            )
        except RobotoDomainException as exc:
            msg = f"Failed to retrieve invocation input data! Reason: {exc.message}"
            log.error(msg)
            is_server_error = exc.http_status_code >= 500
            raise PrepareEnvException(ExitCode.InternalError if is_server_error else ExitCode.UsageError, msg) from None

        # Modify resolved paths to be relative to the target directory
        if requires_downloaded_inputs and download_directory != target_directory:
            resolved_input = _hardlink_resolved_input_to_target_directory(resolved_input, target_directory)

        log.info("Writing input manifest to %s", inputs_data_manifest_file)
        _ensure_file_path_exists(inputs_data_manifest_file)
        inputs_data_manifest_file.write_text(resolved_input.model_dump_json())


def prepare_metadata_changeset_manifest(
    dataset_metadata_changeset_path: pathlib.Path,
):
    log.info(
        "Creating dataset metadata changeset file at %s",
        dataset_metadata_changeset_path,
    )
    _ensure_file_path_exists(dataset_metadata_changeset_path)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import atexit
import collections.abc
import concurrent.futures
import contextlib
import datetime
import logging
import pathlib
import time
import typing
import urllib.request

from ...association import AssociationType
from ...compat import import_optional_dependency
from ...http import RobotoClient
from ...logging import default_logger
from .mcap_reader import McapReader
from .operations import (
    MessagePathRepresentationMapping,
)
from .record import (
    RepresentationRecord,
    RepresentationStorageFormat,
)
from .topic_reader import Timestamp, TopicReader

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep

logger = default_logger()

OUTFILE_NAME_PATTERN = "{repr_id}_{file_id}.mcap"


def garbage_collect_old_topic_data(
    cache_dir: pathlib.Path,
    expire_after: datetime.timedelta = datetime.timedelta(days=7),
):
    pattern_as_glob = OUTFILE_NAME_PATTERN.format(file_id="*", repr_id="*")
    now_in_nanoseconds = time.time_ns()
    expire_after_ns = int(expire_after.total_seconds() * 1_000_000_000)
    for mcap_file in cache_dir.glob(f"**/{pattern_as_glob}"):
        if (now_in_nanoseconds - mcap_file.stat().st_atime_ns) > expire_after_ns:
            mcap_file.unlink()


class McapTopicReader(TopicReader):
    """Private interface for retrieving topic data stored in MCAP files.

    Note:
        This is not intended as a public API.
        To access topic data, prefer the ``get_data`` or ``get_data_as_df`` methods
        on :py:class:`~roboto.domain.topics.Topic`, :py:class:`~roboto.domain.topics.MessagePath`,
        or :py:class:`~roboto.domain.events.Event`.
    """

    __cache_dir: pathlib.Path
    __roboto_client: RobotoClient

    @staticmethod
    def accepts(
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
    ) -> bool:
        for mapping in message_paths_to_representations:
            if mapping.representation.storage_format != RepresentationStorageFormat.MCAP:
                return False
        return True

    def __init__(self, roboto_client: RobotoClient, cache_dir: pathlib.Path):
        self.__roboto_client = roboto_client
        self.__cache_dir = cache_dir

    def get_data(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
        start_time: typing.Optional[int] = None,
        end_time: typing.Optional[int] = None,
        timestamp_message_path_representation_mapping: typing.Optional[MessagePathRepresentationMapping] = None,
    ) -> collections.abc.Generator[tuple[Timestamp, dict[str, typing.Any]], None, None]:
        # Schedule a cleanup of the cache_dir to remove any old assets.
        atexit.register(garbage_collect_old_topic_data, cache_dir=self.__cache_dir)

        repr_id_to_outfile_map = self.__ensure_cached(message_paths_to_representations)

        with contextlib.ExitStack() as exit_stack:
            readers = [
                McapReader(
                    stream=exit_stack.enter_context(
                        repr_id_to_outfile_map[message_path_repr_map.representation.representation_id].open(mode="rb")
                    ),
                    message_paths=message_path_repr_map.message_paths,
                    start_time=start_time,
                    end_time=end_time,
                )
                for message_path_repr_map in message_paths_to_representations
            ]

            if logger.isEnabledFor(logging.DEBUG):
                for reader in readers:
                    logger.debug(
                        "Reader will pick %r message_paths from data",
                        reader.message_paths,
                    )

            while any(reader.has_next for reader in readers):
                full_record = {}
                timestamp = min(reader.next_timestamp for reader in readers)
                for reader in readers:
                    if reader.next_message_is_time_aligned(timestamp):
                        decoded_message = reader.next()
                        if decoded_message is None:
                            continue
                        full_record.update(decoded_message.to_dict())

                yield timestamp, full_record

    def get_data_as_df(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
        start_time: typing.Optional[int] = None,
        end_time: typing.Optional[int] = None,
        timestamp_message_path_representation_mapping: typing.Optional[MessagePathRepresentationMapping] = None,
    ) -> tuple[pandas.Series, pandas.DataFrame]:
        pd = import_optional_dependency("pandas", "analytics")
        timestamps = []
        data = []
        for timestamp, record in self.get_data(
            message_paths_to_representations=message_paths_to_representations,
            start_time=start_time,
            end_time=end_time,
        ):
            timestamps.append(timestamp)
            data.append(record)

        return pd.Series(timestamps), pd.json_normalize(data=data)

    def __ensure_cached(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
    ) -> dict[str, pathlib.Path]:
        repr_id_to_outfile_map: dict[str, pathlib.Path] = {}
        download_list: list[MessagePathRepresentationMapping] = []
        for message_path_repr_map in message_paths_to_representations:
            representation = message_path_repr_map.representation
            outfile = self.__representation_out_file(representation, cache_dir=self.__cache_dir)
            repr_id_to_outfile_map[representation.representation_id] = outfile
            if not outfile.exists():
                association = representation.association
                if association.association_type != AssociationType.File:
                    logger.warning(
                        "Unable to get data for message paths %r",
                        [record.message_path for record in message_path_repr_map.message_paths],
                    )
                    continue
                download_list.append(message_path_repr_map)

        if download_list:
            self.__download_representations(
                message_path_repr_mappings=download_list,
                repr_id_to_outfile_map=repr_id_to_outfile_map,
            )

        return repr_id_to_outfile_map

    def __download_representations(
        self,
        message_path_repr_mappings: collections.abc.Sequence[MessagePathRepresentationMapping],
        repr_id_to_outfile_map: collections.abc.Mapping[str, pathlib.Path],
    ) -> None:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_to_representation_mapping = {
                executor.submit(
                    self.__download_roboto_file,
                    message_path_repr_map.representation,
                    repr_id_to_outfile_map[message_path_repr_map.representation.representation_id],
                ): message_path_repr_map
                for message_path_repr_map in message_path_repr_mappings
            }
            for future in concurrent.futures.as_completed(future_to_representation_mapping):
                message_path_repr_map = future_to_representation_mapping[future]
                try:
                    future.result()
                except Exception:
                    logger.exception(
                        "Unable to get data for message paths %r",
                        [record.message_path for record in message_path_repr_map.message_paths],
                    )

    def __download_roboto_file(
        self,
        representation: RepresentationRecord,
        outfile: pathlib.Path,
    ) -> None:
        file_id = representation.association.association_id
        logger.debug("Getting signed url for file '%s'", file_id)
        signed_url_response = self.__roboto_client.get(f"v1/files/{file_id}/signed-url")
        signed_url = signed_url_response.to_dict(json_path=["data", "url"])

        logger.debug("Downloading file '%s' to %s", file_id, outfile)
        urllib.request.urlretrieve(signed_url, str(outfile))

    def __representation_out_file(
        self,
        representation: RepresentationRecord,
        cache_dir: pathlib.Path,
    ) -> pathlib.Path:
        return cache_dir / OUTFILE_NAME_PATTERN.format(
            repr_id=representation.representation_id,
            file_id=representation.association.association_id,
        )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
import collections
import collections.abc
import typing

import mcap_ros1._vendor.genpy
import mcap_ros2._dynamic

from ...logging import default_logger
from .record import MessagePathRecord

logger = default_logger()


def is_ros1_time_value(val: typing.Any) -> bool:
    return isinstance(val, mcap_ros1._vendor.genpy.TVal)


def is_ros2_time_value(val: typing.Any) -> bool:
    """Structural type checking because Time and Duration classes are dynamically generated"""
    if not hasattr(val, "__slots__"):
        return False

    slots = getattr(val, "__slots__", [])
    for field in mcap_ros2._dynamic.TimeDefinition.fields:
        if field.name not in slots:
            return False

    return True


T = typing.TypeVar("T")


class defaultlist(list[T], typing.Generic[T]):
    """Like collections.defaultdict, but for list.

    Automatically supplies a default value when you access an index that hasn't been set or is out of bounds,
    without raising an IndexError.

    Examples:
        >>> dl = defaultlist[int](factory=lambda: 0)
        >>> dl[5] += 1  # Automatically extends list with 0s up to index 5
        >>> print(dl)  # [0, 0, 0, 0, 0, 1]
    """

    def __init__(self, factory: typing.Callable[[], T]):
        self.factory = factory
        super().__init__()

    @typing.overload
    def __getitem__(self, idx: typing.SupportsIndex) -> T: ...

    @typing.overload
    def __getitem__(self, idx: slice) -> list[T]: ...

    def __getitem__(self, idx: typing.Union[typing.SupportsIndex, slice]) -> typing.Union[T, list[T]]:
        if isinstance(idx, slice):
            return super().__getitem__(idx)
        # Convert SupportsIndex to int for comparison
        index = idx.__index__()
        while index >= len(self):
            self.append(self.factory())
        return super().__getitem__(idx)

    @typing.overload
    def __setitem__(self, idx: typing.SupportsIndex, value: T) -> None: ...

    @typing.overload
    def __setitem__(self, idx: slice, value: typing.Iterable[T]) -> None: ...

    def __setitem__(
        self,
        idx: typing.Union[typing.SupportsIndex, slice],
        value: typing.Union[T, typing.Iterable[T]],
    ) -> None:
        if isinstance(idx, slice):
            super().__setitem__(idx, typing.cast(typing.Iterable[T], value))
        else:
            # Convert SupportsIndex to int for comparison
            index = idx.__index__()
            while index >= len(self):
                self.append(self.factory())
            super().__setitem__(idx, typing.cast(T, value))


class AttrGetter(abc.ABC):
    """Abstract base class for accessing attributes from decoded messages.

    Provides a unified interface for extracting attributes from different types of
    decoded message data, whether they are dictionaries (JSON) or dynamically
    created classes (ROS1/ROS2). Used by message decoders to handle various
    message encoding formats in a consistent way.
    """

    @staticmethod
    @abc.abstractmethod
    def get_attribute_names(value) -> collections.abc.Sequence[str]:
        """Get the names of all attributes available in the given value.

        Args:
            value: The decoded message value to inspect.

        Returns:
            Sequence of attribute names available in the value.
        """

    @staticmethod
    @abc.abstractmethod
    def get_attribute(value, attribute) -> typing.Any:
        """Get the value of a specific attribute from the given value.

        Args:
            value: The decoded message value to access.
            attribute: Name of the attribute to retrieve.

        Returns:
            The value of the specified attribute.
        """

    @staticmethod
    @abc.abstractmethod
    def has_attribute(value, attribute: str) -> bool:
        """Check if the given value has a specific attribute.

        Args:
            value: The decoded message value to inspect.
            attribute: Name of the attribute to check for.

        Returns:
            True if the value has the specified attribute, False otherwise.
        """

    @staticmethod
    @abc.abstractmethod
    def has_sub_attributes(value) -> bool:
        """Check if the given value has nested attributes that can be accessed.

        Args:
            value: The decoded message value to inspect.

        Returns:
            True if the value has nested attributes, False otherwise.
        """


class ClassAttrGetter(AttrGetter):
    """Attribute getter for class-based decoded data.

    Handles access to attributes from decoded messages that are represented as
    dynamically created classes with __slots__ at runtime. This includes ROS1,
    ROS2, and other message formats that use class-based representations.
    """

    @staticmethod
    def get_attribute_names(value):
        return value.__slots__

    @staticmethod
    def get_attribute(value, attribute):
        return getattr(value, attribute)

    @staticmethod
    def has_attribute(value, attribute: str) -> bool:
        return hasattr(value, attribute)

    @staticmethod
    def has_sub_attributes(value):
        return hasattr(value, "__slots__")


class DictAttrGetter(AttrGetter):
    """Attribute getter for JSON decoded data.

    Handles access to attributes from JSON decoded messages, which are
    represented as standard Python dictionaries.
    """

    @staticmethod
    def get_attribute_names(value):
        return value.keys()

    @staticmethod
    def get_attribute(value, attribute):
        return value[attribute]

    @staticmethod
    def has_attribute(value, attribute: str) -> bool:
        return attribute in value

    @staticmethod
    def has_sub_attributes(value):
        return hasattr(value, "keys")


class DecodedMessage:
    """Facade for values returned from message decoders.

    Provides a unified interface for working with decoded messages regardless
    of their original encoding format or source. Handles the conversion of decoded
    message data into dictionary format suitable for analysis and processing.

    A decoded message may be one of several types:
    - A dictionary when the message data is encoded as JSON
    - A dynamically created class when the message data is encoded as ROS1, ROS2 (CDR), or other binary formats

    This class abstracts away the differences between these formats and provides
    consistent access to message data through a dictionary interface, filtering
    the output based on the specified message paths.
    """

    __message: typing.Any
    __message_paths: collections.abc.Sequence[MessagePathRecord]

    @staticmethod
    def is_path_match(attrib: str, message_path: str) -> bool:
        """Check if an attribute path matches or is a parent of a message path.

        Determines whether a given attribute path should be included when filtering
        message data based on the specified message paths.

        Args:
            attrib: Attribute path to check (e.g., "pose.position").
            message_path: Target message path (e.g., "pose.position.x").

        Returns:
            True if the attribute matches or is a parent of the message path.

        Examples:
            >>> DecodedMessage.is_path_match("pose", "pose.position.x")
            True
            >>> DecodedMessage.is_path_match("pose.position", "pose.position.x")
            True
            >>> DecodedMessage.is_path_match("pose.position.x", "pose.position.x")
            True
            >>> DecodedMessage.is_path_match("velocity", "pose.position.x")
            False
        """
        if attrib == message_path:
            return True

        attrib_parts = attrib.split(".")
        path_parts = message_path.split(".")

        if len(attrib_parts) >= len(path_parts):
            return False

        return attrib_parts == path_parts[: len(attrib_parts)]

    def __init__(
        self,
        msg: typing.Any,
        message_paths: collections.abc.Sequence[MessagePathRecord],
    ):
        self.__message = msg
        self.__message_paths = message_paths

    def to_dict(self) -> dict:
        """Convert the decoded message to a dictionary format.

        Extracts and organizes message data into a dictionary structure,
        including only the attributes that match the specified message paths.

        Returns:
            Dictionary containing the filtered message data with attribute names as keys.

        Examples:
            >>> # Assuming message_paths include "pose.position.x" and "velocity"
            >>> decoded_msg = DecodedMessage(ros_message, message_paths)
            >>> data_dict = decoded_msg.to_dict()
            >>> print(data_dict)
            {'pose': {'position': {'x': 1.5}}, 'velocity': 2.0}
        """
        accumulator: dict[str, typing.Any] = {}

        getter: AttrGetter = ClassAttrGetter()

        if isinstance(self.__message, dict):
            # Message data was encoded as JSON
            getter = DictAttrGetter()

        for record in self.__message_paths:
            self.__extract_path(
                accumulator=accumulator,
                obj=self.__message,
                path_components=record.path_in_schema.copy(),
                path_index=0,
                getter=getter,
            )

        return accumulator

    def __extract_path(
        self,
        accumulator: dict[str, typing.Any],
        obj: typing.Any,
        path_components: list[str],
        path_index: int,
        getter: AttrGetter,
    ) -> None:
        """Recursively extract values from a nested object structure following a path.

        Traverses the object using path_components and builds a nested dictionary in the
        accumulator. Handles special cases for ROS1/ROS2 time fields and sequences.

        Args:
            accumulator: Dictionary to store the extracted values. Modified in place.
            obj: The current object being traversed.
            path_components: List of attribute names forming the path to extract.
            path_index: Current position in the path_components list (0-based).
            getter: AttrGetter instance for accessing attributes from obj.
        """
        # Base case: we've exhausted the path components
        if path_index >= len(path_components):
            return

        current_attr = path_components[path_index]

        # Check if the attribute exists in the object
        if not getter.has_attribute(obj, current_attr):
            return

        value = getter.get_attribute(obj, current_attr)

        # Leaf node: this is the final attribute in the path
        if path_index == len(path_components) - 1:
            accumulator[current_attr] = value
            return

        if is_ros1_time_value(value):
            for canonical_msg_path_name, attr_name in [
                # ROS1 uses "secs"
                #   https://github.com/foxglove/mcap/blob/main/python/mcap-ros1-support/mcap_ros1/_vendor/genpy/rostime.py#L54
                # and "nsecs"
                #   https://github.com/foxglove/mcap/blob/main/python/mcap-ros1-support/mcap_ros1/_vendor/genpy/rostime.py#L54
                # but ROS ingestion uses "sec" and "nsec" instead for message path naming:
                #   https://github.com/roboto-ai/roboto-ingestion-actions/blob/main/actions/ros_ingestion/src/ros_ingestion/utils.py#L408-L420
                ("sec", "secs"),
                ("nsec", "nsecs"),
            ]:
                try:
                    index_of_sec = path_components.index(canonical_msg_path_name)
                    if index_of_sec > path_index:
                        path_components[index_of_sec] = attr_name
                except ValueError:
                    pass
        elif is_ros2_time_value(value):
            # ROS2 uses "nanosec"
            #   https://github.com/foxglove/mcap/blob/main/python/mcap-ros2-support/mcap_ros2/_dynamic.py#L131
            # but ROS ingestion uses "nsec" instead for message path naming:
            #   https://github.com/roboto-ai/roboto-ingestion-actions/blob/main/actions/ros_ingestion/src/ros_ingestion/utils.py#L401-L402
            try:
                index_of_sec = path_components.index("nsec")
                if index_of_sec > path_index:
                    path_components[index_of_sec] = "nanosec"
            except ValueError:
                pass
        elif isinstance(value, collections.abc.Sequence) and not isinstance(value, (str, bytes)):
            # Map over the list and extract from each element
            sub_path = path_components[path_index + 1 :]
            # Merge with existing list results
            list_accum = (
                accumulator[current_attr]
                if current_attr in accumulator and isinstance(accumulator[current_attr], list)
                else defaultlist[dict](factory=dict)
            )
            for i, item in enumerate(value):
                item_accum = list_accum[i]
                self.__extract_path(
                    accumulator=item_accum,
                    obj=item,
                    path_components=sub_path,
                    path_index=0,
                    getter=getter,
                )

            accumulator[current_attr] = list_accum
            return

        if not getter.has_sub_attributes(value):
            # Can't traverse further, but we haven't reached the end of the indexable path
            return

        # Create nested obj if it doesn't exist
        if current_attr not in accumulator:
            accumulator[current_attr] = dict()

        # Recurse to the next level
        self.__extract_path(
            accumulator=accumulator[current_attr],
            obj=value,
            path_components=path_components,
            path_index=path_index + 1,
            getter=getter,
        )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum
import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:

    class StrEnum(str, enum.Enum):
        """Compatibility StrEnum for Python < 3.11"""

        def __str__(self) -> str:
            return self.value



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import pathlib
import typing

from ..association import Association
from ..http import RobotoClient
from .download_session import (
    DownloadableFile,
    DownloadSession,
)
from .object_store import (
    OnProgress,
    StoreRegistry,
)
from .upload_transaction import (
    TransactionFile,
    UploadTransaction,
)

_DEFAULT_UPLOAD_BATCH_SIZE = 500


class FileService:
    """Application service for performing upload and download to the Roboto Platform.

    Agnostic to object store provider.
    """

    def __init__(
        self,
        roboto_client: typing.Optional[RobotoClient] = None,
        object_store_registry: typing.Optional[StoreRegistry] = None,
    ):
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__object_store_registry = object_store_registry or StoreRegistry

    def upload(
        self,
        files: collections.abc.Iterable[pathlib.Path],
        association: Association,
        destination_paths: collections.abc.Mapping[pathlib.Path, str] = {},
        batch_size: int = _DEFAULT_UPLOAD_BATCH_SIZE,
        device_id: typing.Optional[str] = None,
        caller_org_id: typing.Optional[str] = None,
        on_progress: typing.Optional[OnProgress] = None,
    ) -> list[str]:
        items: list[TransactionFile] = []
        for local_path in files:
            try:
                file_size = local_path.stat().st_size
            except OSError as e:
                raise OSError(f"Cannot upload file '{local_path}': {str(e)}") from None

            items.append(
                {
                    "local_path": local_path,
                    "destination_path": destination_paths.get(local_path, local_path.name),
                    "file_size": file_size,
                }
            )

        if not items:
            return []

        # GM(2025-11-19)
        # For reasons related to OpenFGA scalability/throughput,
        # upload transactions are currently limited to 500 files.
        # Until that is fixed, implement batching by creating multiple transactions.
        # When that is lifted, batching is already handled by the UploadTransaction.
        completed_upload_node_ids: list[str] = []
        for batch_start in range(0, len(items), batch_size):
            item_batch = items[batch_start : batch_start + batch_size]

            with UploadTransaction(
                item_batch,
                association,
                device_id=device_id,
                batch_size=batch_size,
                roboto_client=self.__roboto_client,
                caller_org_id=caller_org_id,
            ) as txn:
                # Heuristic: all files in a transaction are located in the same object store.
                # When this no longer holds, this is the place to change it.
                first_file_uri = list(txn.upload_mappings.values())[0]

                object_store = self.__object_store_registry.get_store_for_uri(
                    first_file_uri, txn.make_credential_provider()
                )
                with object_store:
                    for file in txn:
                        future = object_store.put(file["local_path"], file["upload_uri"], on_progress=on_progress)
                        txn.register_upload(file, future)

                completed_upload_node_ids.extend(txn.completed_upload_node_ids)

        return completed_upload_node_ids

    def download(
        self,
        files: collections.abc.Sequence[DownloadableFile],
        association: Association,
        caller_org_id: typing.Optional[str] = None,
        on_progress: typing.Optional[OnProgress] = None,
    ) -> None:
        """Download files from the Roboto Platform.

        Args:
            files: Sequence of files to download, each with source_uri and destination_path.
            association: Association of the files to download.
            caller_org_id: Optional organization ID for cross-org access.
            on_progress: Optional callback to be periodically called with the number of bytes downloaded.
        """
        if not files:
            return

        files_grouped_by_bucket = collections.defaultdict(list)
        for file in files:
            files_grouped_by_bucket[file["bucket_name"]].append(file)

        for bucket_name, bucket_files in files_grouped_by_bucket.items():
            download_session = DownloadSession(
                items=bucket_files,
                association=association,
                roboto_client=self.__roboto_client,
                caller_org_id=caller_org_id,
            )
            # Heuristic: all files in the same bucket are located in the same object store
            first_file_uri = bucket_files[0]["source_uri"]
            object_store = self.__object_store_registry.get_store_for_uri(
                first_file_uri, download_session.make_credential_provider(bucket_name)
            )

            with object_store, download_session:
                for file in bucket_files:
                    future = object_store.get(file["source_uri"], file["destination_path"], on_progress=on_progress)
                    download_session.register_download(file, future)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

from ....compat import import_optional_dependency
from ....logging import default_logger
from ..operations import AddMessagePathRequest
from ..record import (
    CanonicalDataType,
    MessagePathMetadataWellKnown,
    MessagePathStatistic,
)
from .parquet_parser import ParquetParser
from .timestamp import TimestampInfo

if typing.TYPE_CHECKING:
    import pyarrow  # pants: no-infer-dep


logger = default_logger()


def arrow_type_to_canonical_type(
    arrow_type: "pyarrow.DataType",
) -> CanonicalDataType:
    pa = import_optional_dependency("pyarrow", "ingestion")

    if pa.types.is_boolean(arrow_type):
        return CanonicalDataType.Boolean

    if pa.types.is_integer(arrow_type) or pa.types.is_decimal(arrow_type) or pa.types.is_floating(arrow_type):
        return CanonicalDataType.Number

    if pa.types.is_timestamp(arrow_type) and typing.cast("pyarrow.TimestampType", arrow_type).tz == "UTC":
        return CanonicalDataType.Timestamp

    if pa.types.is_string(arrow_type) or pa.types.is_large_string(arrow_type):
        return CanonicalDataType.String

    if pa.types.is_dictionary(arrow_type):
        # Interpret a DictionaryArray as "categorical" data.
        # PyArrow's ability to auto-interpret a Parquet column as a DictionaryArray depends on
        # saving the Arrow schema in the Parquet file's FileMetadata.
        # Else, the ingestion action would need to know ahead of time which columns to parse as dictionaries
        # (i.e., via the `read_dictionary` __init__ arg to ParquetFile).
        return CanonicalDataType.Categorical

    if pa.types.is_map(arrow_type) or pa.types.is_struct(arrow_type):
        return CanonicalDataType.Object

    if pa.types.is_list(arrow_type) or pa.types.is_large_list(arrow_type):
        value_type = typing.cast("pyarrow.ListType", arrow_type).value_type
        # list<numeric> -> NumberArray, list<non-numeric> or list<struct> -> Array
        if pa.types.is_integer(value_type) or pa.types.is_floating(value_type) or pa.types.is_decimal(value_type):
            return CanonicalDataType.NumberArray
        return CanonicalDataType.Array

    if pa.types.is_binary(arrow_type) or pa.types.is_large_binary(arrow_type):
        return CanonicalDataType.Byte

    return CanonicalDataType.Unknown


def sanitize_column_name(field: "pyarrow.Field") -> str:
    return field.name.replace(".", "_")


def compute_numeric_statistics(
    data: typing.Union["pyarrow.Array", "pyarrow.ChunkedArray"],
) -> dict[str, typing.Any]:
    pc = import_optional_dependency("pyarrow.compute", "ingestion")

    min_max = pc.min_max(data)
    mean = pc.mean(data)
    # GM(2025-04-20)
    # PyArrow does not implement a true median function.
    # It instead provides `pc.approximate_median`, which uses T-Digest under the hood.
    # That algorithm is optimized for accuracy at the tails of a distribution,
    # trading off for accuracy determining central tendency.
    # The primary knob we have to control is to set the compression parameter, which defaults (in PyArrow) to 100.
    # A higher value forces the algorithm to create and maintain more centroids,
    # which lead to finer-grained summaries and generally higher accuracy.
    # The number of centroids, and thus the memory footprint of the digest,
    # is roughly proportional to the compression value.
    # Higher accuracy comes at the cost of increased resource consumption.
    # Two other factors affect the accuracy:
    #   1. Size of data: larger series should produce more accurate estimates
    #   2. Data distribution: the accuracy is not uniform across all distrubtion shapes
    median = pc.tdigest(
        data,
        # median
        q=0.5,
        # See note above about choice of significantly larger compression value
        # Research by Apache Presto indicates a compression of 500 achieves very low error median approximations.
        # https://github.com/prestodb/presto/issues/12929
        delta=500,
    )
    return {
        MessagePathStatistic.Min.value: min_max["min"].as_py(),
        MessagePathStatistic.Max.value: min_max["max"].as_py(),
        MessagePathStatistic.Mean.value: mean.as_py(),
        MessagePathStatistic.Median.value: median[0].as_py(),
    }


def compute_boolean_statistics(
    data: typing.Union["pyarrow.Array", "pyarrow.ChunkedArray"],
) -> dict[str, typing.Any]:
    pc = import_optional_dependency("pyarrow.compute", "ingestion")

    non_null_count = pc.count(data, mode="only_valid").as_py()
    if non_null_count == 0:
        return {"true_count": 0, "false_count": 0}
    true_count = pc.sum(data).as_py()  # True == 1, so sum() is count of True
    false_count = non_null_count - true_count
    return {"true_count": true_count, "false_count": false_count}


def compute_dictionary_metadata(
    column_name: str,
    data: typing.Union["pyarrow.Array", "pyarrow.ChunkedArray"],
    max_dictionary_size: int = 2048,  # bytes (2kB; arbitrarily chosen)
) -> dict[str, typing.Any]:
    pa = import_optional_dependency("pyarrow", "ingestion")

    arr = typing.cast("pyarrow.ChunkedArray", data).combine_chunks() if isinstance(data, pa.ChunkedArray) else data
    if not isinstance(arr, pa.DictionaryArray):
        return dict()

    dictionary = typing.cast("pyarrow.DictionaryArray", arr).dictionary
    categories = dictionary.to_pylist()

    # Ensure the size of the categories list won't explode the size of the MessagePathRecord
    combined_byte_length = 0
    for category in categories:
        value = category.encode() if isinstance(category, str) else category
        size = len(value) if isinstance(value, collections.abc.Sized) else 0
        combined_byte_length += size

    if combined_byte_length > max_dictionary_size:
        logger.warning(
            "'%s': categories list is larger (%d bytes) than the allowed maximum (%d bytes). "
            "While this column will still be inferred as containing 'categorical' data, "
            "the Roboto Platform will be unable to map the categories to their integer values.",
            column_name,
            combined_byte_length,
            max_dictionary_size,
        )
        return dict()

    return {MessagePathMetadataWellKnown.Categories.value: categories}


def get_nested_column_data(
    parser: ParquetParser,
    column_name: str,
    field_path: list[str],
) -> typing.Union["pyarrow.Array", "pyarrow.ChunkedArray"]:
    """Extract data for nested fields from a PyArrow table.

    Navigates through struct fields using the provided field path to extract
    the data for a nested field.

    Args:
        parser: ParquetParser instance to read data from.
        column_name: Name of the top-level column.
        field_path: List of field names to traverse (excluding the column name).

    Returns:
        The extracted Array or ChunkedArray for the nested field.

    Raises:
        KeyError: If a field in the path does not exist.
    """
    pc = import_optional_dependency("pyarrow.compute", "ingestion")

    data = parser.get_data_for_column(column_name).column(0)

    for component in field_path:
        # Use pc.struct_field to extract nested struct fields from ChunkedArray
        data = pc.struct_field(data, component)

    return data


def get_list_element_data(
    parser: ParquetParser,
    column_name: str,
    field_path: list[str],
) -> typing.Union["pyarrow.Array", "pyarrow.ChunkedArray"]:
    """Extract flattened data from list columns for statistics computation.

    For list<primitive> columns, flattens all list elements into a single array.
    For list<struct> columns, flattens and then accesses the struct field.

    Args:
        parser: ParquetParser instance to read data from.
        column_name: Name of the top-level column.
        field_path: List of field names to traverse after flattening the list.

    Returns:
        The flattened Array or ChunkedArray suitable for statistics computation.
    """
    pc = import_optional_dependency("pyarrow.compute", "ingestion")

    data = parser.get_data_for_column(column_name).column(0)

    # Flatten the list to get all elements
    data = pc.list_flatten(data)

    # If there are field path components, navigate into struct fields
    for component in field_path:
        data = pc.struct_field(data, component)

    return data


def compute_field_metadata(
    parser: ParquetParser,
    column_name: str,
    field_path: list[str],
    canonical_data_type: CanonicalDataType,
    is_inside_list: bool = False,
) -> dict[str, typing.Any]:
    """Compute metadata including statistics for a field.

    Handles both top-level and nested fields, extracting data appropriately
    based on the field's location in the schema hierarchy.

    Args:
        parser: ParquetParser instance to read data from.
        column_name: Name of the top-level column.
        field_path: List of field names to traverse (empty for top-level fields).
        canonical_data_type: The canonical type of the field.
        is_inside_list: Whether this field is inside a list (affects data extraction).

    Returns:
        Metadata dictionary with statistics if applicable.
    """
    pc = import_optional_dependency("pyarrow.compute", "ingestion")

    full_path = ".".join([column_name] + field_path)
    metadata: dict[str, typing.Any] = {}

    # Determine which types need data extraction for statistics
    types_needing_stats = (
        CanonicalDataType.Number,
        CanonicalDataType.NumberArray,
        CanonicalDataType.Boolean,
        CanonicalDataType.Categorical,
    )
    if canonical_data_type not in types_needing_stats:
        return metadata

    try:
        # Extract data based on field location
        if is_inside_list:
            # Field inside a list<struct> - flatten and extract
            data = get_list_element_data(parser, column_name, field_path)
        else:
            data = get_nested_column_data(parser, column_name, field_path)
            if canonical_data_type == CanonicalDataType.NumberArray:
                data = pc.list_flatten(data)

        # Compute type-specific statistics
        if canonical_data_type in (CanonicalDataType.Number, CanonicalDataType.NumberArray):
            metadata.update(compute_numeric_statistics(data))
        elif canonical_data_type == CanonicalDataType.Boolean:
            metadata.update(compute_boolean_statistics(data))
        elif canonical_data_type == CanonicalDataType.Categorical:
            metadata.update(compute_dictionary_metadata(full_path, data))

    except Exception as e:
        logger.warning(
            "Failed to compute statistics for nested field '%s': %s",
            full_path,
            str(e),
        )

    return metadata


def _traverse_field(
    field: "pyarrow.Field",
    path_prefix: str,
    field_path: list[str],
    column_name: str,
    parser: ParquetParser,
    timestamp: TimestampInfo,
    depth: int,
    max_depth: int,
    is_inside_list: bool,
) -> typing.Generator[AddMessagePathRequest, None, None]:
    """Recursively traverse a field and yield AddMessagePathRequest objects.

    Args:
        field: The PyArrow field to traverse.
        path_prefix: The dot-delimited path prefix for this field.
        field_path: List of field names from the column to this field (for data extraction).
        column_name: The top-level column name (for data extraction).
        parser: ParquetParser instance.
        timestamp: Timestamp information.
        depth: Current recursion depth.
        max_depth: Maximum recursion depth.
        is_inside_list: Whether we are inside a list type (affects child type mapping).

    Yields:
        AddMessagePathRequest objects for this field and its children.
    """
    pa = import_optional_dependency("pyarrow", "ingestion")

    # Build the message path for this field
    field_name_sanitized = field.name.replace(".", "_")
    message_path = f"{path_prefix}.{field_name_sanitized}" if path_prefix else field_name_sanitized

    # Determine canonical type
    arrow_type = field.type
    canonical_data_type = arrow_type_to_canonical_type(arrow_type)

    # Special handling for timestamp field
    if field == timestamp.field:
        canonical_data_type = CanonicalDataType.Timestamp

    # For fields inside a list<struct>, numeric types stay as Number (not NumberArray)
    # because the list parent already indicates the array nature

    # Build path_in_schema for this field
    full_field_path = [column_name] + field_path

    # Handle struct types - yield parent as Object, then recurse into children
    if pa.types.is_struct(arrow_type):
        yield AddMessagePathRequest(
            canonical_data_type=CanonicalDataType.Object,
            data_type=str(arrow_type),
            message_path=message_path,
            metadata={},
            path_in_schema=full_field_path,
        )

        # Recurse into struct fields if within depth limit
        if depth < max_depth:
            struct_type = typing.cast("pyarrow.StructType", arrow_type)
            for i in range(struct_type.num_fields):
                child_field = struct_type.field(i)
                child_field_path = field_path + [child_field.name]
                yield from _traverse_field(
                    field=child_field,
                    path_prefix=message_path,
                    field_path=child_field_path,
                    column_name=column_name,
                    parser=parser,
                    timestamp=timestamp,
                    depth=depth + 1,
                    max_depth=max_depth,
                    is_inside_list=is_inside_list,
                )
        else:
            logger.warning(
                "'%s' has nested data beyond the maximum supported depth (%d). "
                "Its child fields will not be individually searchable or plottable, "
                "and will instead appear as a raw nested object.",
                message_path,
                max_depth,
            )
        return

    # Handle list types
    if pa.types.is_list(arrow_type) or pa.types.is_large_list(arrow_type):
        value_type = typing.cast("pyarrow.ListType", arrow_type).value_type

        # Yield the list field itself
        metadata = compute_field_metadata(
            parser=parser,
            column_name=column_name,
            field_path=field_path,
            canonical_data_type=canonical_data_type,
            is_inside_list=is_inside_list,
        )
        yield AddMessagePathRequest(
            canonical_data_type=canonical_data_type,
            data_type=str(arrow_type),
            message_path=message_path,
            metadata=metadata,
            path_in_schema=full_field_path,
        )

        # For list<struct>, recurse into struct fields
        if pa.types.is_struct(value_type):
            if depth < max_depth:
                struct_type = typing.cast("pyarrow.StructType", value_type)
                for i in range(struct_type.num_fields):
                    child_field = struct_type.field(i)
                    child_field_path = field_path + [child_field.name]
                    yield from _traverse_field(
                        field=child_field,
                        path_prefix=message_path,
                        field_path=child_field_path,
                        column_name=column_name,
                        parser=parser,
                        timestamp=timestamp,
                        depth=depth + 1,
                        max_depth=max_depth,
                        is_inside_list=True,  # Mark that we're inside a list
                    )
            else:
                logger.warning(
                    "'%s' has nested data beyond the maximum supported depth (%d). "
                    "Its child fields will not be individually searchable or plottable, "
                    "and will instead appear as a raw nested object.",
                    message_path,
                    max_depth,
                )
        return

    metadata = compute_field_metadata(
        parser=parser,
        column_name=column_name,
        field_path=field_path,
        canonical_data_type=canonical_data_type,
        is_inside_list=is_inside_list,
    )

    # Special handling for timestamp field metadata
    if field == timestamp.field:
        metadata[MessagePathMetadataWellKnown.Unit.value] = str(timestamp.unit)

    yield AddMessagePathRequest(
        canonical_data_type=canonical_data_type,
        data_type=str(arrow_type),
        message_path=message_path,
        metadata=metadata,
        path_in_schema=full_field_path,
    )


def generate_message_path_requests(
    parser: ParquetParser,
    timestamp: TimestampInfo,
    max_depth: int = 10,
) -> typing.Generator[AddMessagePathRequest, None, None]:
    """Generate AddMessagePathRequest objects for all fields in a Parquet schema.

    Traverses the schema recursively to generate message paths for nested types
    (structs, lists) in addition to top-level fields.

    Args:
        parser: ParquetParser instance containing the schema and data.
        timestamp: Timestamp information for the topic.
        max_depth: Maximum recursion depth for nested types (default: 10).

    Yields:
        AddMessagePathRequest objects for each field and nested field in the schema.

    Examples:
        For a schema with a struct column `position: struct<x: float, y: float>`:
        - Yields `position` (Object)
        - Yields `position.x` (Number)
        - Yields `position.y` (Number)

        For a schema with `values: list<float64>`:
        - Yields `values` (NumberArray)

        For a schema with `points: list<struct<x: float, y: float>>`:
        - Yields `points` (Array)
        - Yields `points.x` (Number)
        - Yields `points.y` (Number)
    """
    for field in parser.fields:
        yield from _traverse_field(
            field=field,
            path_prefix="",
            field_path=[],
            column_name=field.name,
            parser=parser,
            timestamp=timestamp,
            depth=0,
            max_depth=max_depth,
            is_inside_list=False,
        )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import typing

from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Prompt
from rich.spinner import Spinner
from rich.table import Table
from rich.text import Text

from ...ai.chat import Chat
from ...ai.chat.event import (
    ChatStartTextEvent,
    ChatTextDeltaEvent,
    ChatTextEndEvent,
    ChatToolResultEvent,
    ChatToolUseEvent,
)
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext

VALID_EXIT_COMMANDS = ("quit", "exit", "q", "e")


class RobotoPrompt(Prompt):
    prompt_suffix = "> "


class RobotoSpinner:
    def __init__(self, console: Console, text: str, transient: bool = False):
        spinner = Spinner("dots", text=text, style="cyan")
        self._text = text
        self._live = Live(spinner, console=console, refresh_per_second=10, transient=transient)

    def start(self):
        self._live.start()

    def stop(self):
        # Use Text.from_markup to properly render rich markup in the text
        self._live.update(Text.from_markup(f"[bold green]✓ {self._text}[/bold green]"))
        self._live.stop()


class RobotoMultiToolSpinner:
    def __init__(self, console: Console, transient: bool = False):
        self._console = console
        self._live = Live(console=console, refresh_per_second=10, transient=transient)
        self._spinners: dict[str, RobotoSpinner] = {}

        self._in_progress_tools: set[str] = set()
        self._completed_tools: set[str] = set()
        self._failed_tools: set[str] = set()
        self._tool_id_to_name: dict[str, str] = {}

    def begin_tool_call(self, event: ChatToolUseEvent):
        self._tool_id_to_name[event.tool_use_id] = event.name
        self._in_progress_tools.add(event.tool_use_id)
        self._render()

    def end_tool_call(self, event: ChatToolResultEvent):
        self._in_progress_tools.remove(event.tool_use_id)
        if event.success:
            self._completed_tools.add(event.tool_use_id)
        else:
            self._failed_tools.add(event.tool_use_id)
        self._render()

    def has_in_progress_tools(self) -> bool:
        return len(self._in_progress_tools) > 0

    def _render(self):
        grid = Table.grid()
        grid.add_column()

        for tool_id in self._completed_tools:
            grid.add_row(Text.from_markup(f"[bold green]✓ {self._tool_id_to_name[tool_id]}[/bold green]"))

        for tool_id in self._failed_tools:
            grid.add_row(Text.from_markup(f"[bold red]✗ {self._tool_id_to_name[tool_id]}[/bold red]"))

        for tool_id in self._in_progress_tools:
            grid.add_row(
                Spinner(
                    "dots",
                    text=f"Using tool [bold]{self._tool_id_to_name[tool_id]}[/bold]",
                    style="cyan",
                )
            )

        self._live.update(grid)

    def start(self):
        self._live.start()

    def stop(self):
        self._failed_tools.update(self._in_progress_tools)
        self._in_progress_tools.clear()
        self._render()
        self._live.stop()


class RobotoMarkdown:
    def __init__(self, content: str, console: Console):
        self._console = console
        self._content = content
        md = Markdown(content)
        self._live = Live(md, console=console, refresh_per_second=4)

    def start(self):
        self._live.start()

    def add_content(self, content: str):
        self._content += content

        if len(self._content) > 1000 and "\n\n" in self._content:
            break_idx = self._content.rfind("\n\n")
            original_content = self._content[: break_idx + 2]
            new_content = self._content[break_idx + 2 :]
            self._live.update(Markdown(original_content))
            self._live.stop()
            self._content = new_content
            self._live = Live(Markdown(self._content), console=self._console, refresh_per_second=4)
            self._live.start()

        md = Markdown(self._content)
        self._live.update(md)

    def stop(self):
        self._live.stop()


def _print_welcome_banner(console: Console):
    """Print a stylized welcome banner using rich components."""
    # Create ASCII art with rich styling
    art_lines = [
        "      [white]@@@@@@@@@@[/white]        ",
        "   [white]@@@@@@@@@@@@@@@@[/white]     ",
        "  [white]@@@@@@@@@@@@@@@@@@@[/white]   ",
        " [white]@@@@@@        @@@@@@@[/white]  ",
        "[white]@@@@@    [red]====[white]    @@@@@[/white]  ",
        "[red]        ======[white]   @@@@@[/white]  ",
        "[red]        ======[white]   @@@@@[/white]  ",
        "[white]@@@@@    [red]====[white]    @@@@@[/white]  ",
        "[white]@@@@@          @@@@@@[/white]   ",
        "[white]@@@@@       @@@@@@@@[/white]    ",
        "[white]@@@@@ [blue]     [white]@@@@@@@[/white]      ",
        "[white]@@@@@ [blue]oboto [white]@@@@@@[/white]      ",
        "[white]@@@@@ [blue]chat   [white]@@@@@@[/white]     ",
        "[white]@@@@@ [blue]client  [white]@@@@@@[/white]    ",
        "[white]@@@@@ [blue]for cool  [white]@@@@@[/white]   ",
        "[white]@@@@@ [blue]engineers  [white]@@@@@@[/white] ",
    ]

    art_text = "\n".join(art_lines)
    panel = Panel(
        art_text,
        border_style="cyan",
        padding=(1, 2),
        title="[bold cyan]Roboto Chat[/bold cyan]",
        subtitle="[dim]AI-powered assistant[/dim]",
    )
    console.print(panel)
    console.print()


def start(args, context: CLIContext, parser: argparse.ArgumentParser):
    org_id = args.org

    console = Console()
    _print_welcome_banner(console)

    console.print(
        f"[dim]What do you want to chat about? (Enter [cyan]{'/'.join(VALID_EXIT_COMMANDS)}[/cyan] to exit)[/dim]"
    )
    console.print()

    chat: typing.Union[Chat, None] = None
    thinking_spinner: RobotoSpinner | None
    multi_tool_spinner: RobotoMultiToolSpinner | None = None
    markdown: RobotoMarkdown | None = None

    while True:
        try:
            prompt = RobotoPrompt.ask(console=console)
        except (KeyboardInterrupt, EOFError):
            console.print("\n[yellow]Goodbye![/yellow]")
            break

        if prompt.lower() in VALID_EXIT_COMMANDS:
            console.print("[yellow]Goodbye![/yellow]")
            break

        if prompt.strip() == "":
            continue

        console.print()
        thinking_spinner = RobotoSpinner(console=console, text="Thinking...", transient=True)
        thinking_spinner.start()

        if chat is None:
            chat = Chat.start(
                message=prompt,
                org_id=org_id,
                roboto_client=context.roboto_client,
            )
        else:
            chat.send_text(prompt)

        for event in chat.stream_events():
            if thinking_spinner is not None:
                thinking_spinner.stop()
                thinking_spinner = None

            if isinstance(event, ChatStartTextEvent):
                if multi_tool_spinner is not None:
                    multi_tool_spinner.stop()
                    multi_tool_spinner = None
                    console.print()

                markdown = RobotoMarkdown(content="", console=console)
                markdown.start()

            elif isinstance(event, ChatTextDeltaEvent):
                if markdown is not None:
                    markdown.add_content(event.text)

            elif isinstance(event, ChatTextEndEvent):
                if markdown is not None:
                    markdown.stop()
                    console.print()

            elif isinstance(event, ChatToolUseEvent):
                if multi_tool_spinner is None:
                    multi_tool_spinner = RobotoMultiToolSpinner(console=console)
                    multi_tool_spinner.start()

                multi_tool_spinner.begin_tool_call(event)

            elif isinstance(event, ChatToolResultEvent):
                if multi_tool_spinner is not None:
                    multi_tool_spinner.end_tool_call(event)
                    if not multi_tool_spinner.has_in_progress_tools():
                        multi_tool_spinner.stop()
                        multi_tool_spinner = None
                        console.print()

                        thinking_spinner = RobotoSpinner(console=console, text="Thinking...", transient=True)
                        thinking_spinner.start()

        if thinking_spinner is not None:
            thinking_spinner.stop()


def start_setup_parser(parser):
    add_org_arg(parser)


start_command = RobotoCommand(
    name="start",
    logic=start,
    setup_parser=start_setup_parser,
    command_kwargs={"help": "Start an interactive chat session."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import logging
import os
import pathlib
import sys
from typing import Optional

from packaging.version import InvalidVersion, Version
import pydantic

from ..http import HttpClient
from ..logging import LOGGER_NAME
from ..time import utcnow
from ..version import __version__
from .terminal import AnsiColor

logger = logging.getLogger(LOGGER_NAME)


GITHUB_RELEASES_URL = "https://api.github.com/repos/roboto-ai/roboto-python-sdk/releases/latest"


class CLIState(pydantic.BaseModel):
    last_checked_version: Optional[datetime.datetime] = None
    last_version: str = "0.0.0"
    out_of_date: bool = True


def _get_latest_version_from_github() -> Optional[str]:
    """
    Fetch the latest release version from GitHub Releases API.

    Returns:
        The latest version string (without 'v' prefix), or None if the request fails.
    """
    http = HttpClient()

    try:
        headers = {
            "Accept": "application/vnd.github+json",
            "User-Agent": f"roboto-cli/{__version__} (+https://github.com/roboto-ai)",
            "X-GitHub-Api-Version": "2022-11-28",
        }

        response = http.get(GITHUB_RELEASES_URL, headers=headers)
        release_data = response.to_dict()

        tag_name = release_data.get("tag_name")
        if not tag_name:
            logger.debug("GitHub release response missing 'tag_name' field")
            return None

        version = tag_name.lstrip("v")

        return version
    except Exception as exc:
        # Version checking should never block CLI usage
        logger.debug("Failed to fetch latest version from GitHub: %s", exc)
        return None


def is_version_outdated(current: str, latest: str) -> bool:
    """
    Check if the current version is older than the latest version.

    Args:
        current: The current version string (e.g., "1.0.0").
        latest: The latest available version string (e.g., "1.1.0").

    Returns:
        True if current version is older than latest, False otherwise.
        Returns False if either version string is invalid.
    """
    try:
        return Version(current) < Version(latest)
    except InvalidVersion:
        logger.debug("Invalid version format: current=%s, latest=%s", current, latest)
        return False


def check_last_update():
    roboto_tmp_dir = pathlib.Path.home() / ".roboto" / "tmp"
    roboto_tmp_dir.mkdir(parents=True, exist_ok=True)
    cli_state_file = roboto_tmp_dir / "cli_state.json"

    last_version = None

    state = CLIState(last_checked_version=None)
    if cli_state_file.is_file():
        state = CLIState.model_validate_json(cli_state_file.read_text())
        last_version = state.last_version

    if (
        state.last_checked_version is None
        or __version__ != last_version
        or state.out_of_date is None
        or state.out_of_date is True
        or (utcnow() - datetime.timedelta(hours=1)) > state.last_checked_version
    ):
        latest = _get_latest_version_from_github()
        if latest is None:
            return

        state.last_checked_version = utcnow()
        state.last_version = __version__
        state.out_of_date = is_version_outdated(__version__, latest)

        cli_state_file.write_text(state.model_dump_json())

        suppress_message = os.getenv("ROBOTO_CLI_SUPPRESS_UPGRADE_PROMPT", "false").lower() != "false"

        if state.out_of_date and not suppress_message:
            notice = f"{AnsiColor.BLUE}[notice]{AnsiColor.END}"
            print(
                f"\n{notice} A new release of roboto is available: "
                + f"{AnsiColor.RED + __version__ + AnsiColor.END} -> {AnsiColor.GREEN + latest + AnsiColor.END}\n"
                + f"{notice} To update, follow Upgrade CLI instructions at "
                + "https://github.com/roboto-ai/roboto-python-sdk/blob/main/README.md",
                file=sys.stderr,
            )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import sys

from ..command import RobotoCommand
from ..context import CLIContext


def get_directory_size(path) -> int:
    """Calculate the total size of a directory in bytes."""
    total = 0
    for entry in path.rglob("*"):
        if entry.is_file():
            total += entry.stat().st_size
    return total


def size(args, context: CLIContext, parser: argparse.ArgumentParser):
    cache_dir = context.roboto_config.get_cache_dir()

    if not cache_dir.exists():
        print(f"Cache directory does not exist: {cache_dir}", file=sys.stderr)
        print(0)
        return

    total_bytes = get_directory_size(cache_dir)
    print(total_bytes)


def size_setup_parser(parser):
    pass


size_command = RobotoCommand(
    name="size",
    logic=size,
    setup_parser=size_setup_parser,
    command_kwargs={"help": "Returns the total size in bytes of the Roboto cache directory."},
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import pathlib
import sys
import types
import typing

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

from ..association import Association
from ..env import RobotoEnv
from ..exceptions import RobotoInternalException
from ..http import RobotoClient
from ..logging import default_logger
from ..version import roboto_version
from .api_operations import BeginUploadRequest, BeginUploadResponse, ReportUploadProgressRequest
from .credentials import RobotoCredentials
from .object_store import (
    CredentialProvider,
    Credentials,
    FutureLike,
)

logger = default_logger()


class TransactionFile(typing.TypedDict):
    local_path: pathlib.Path
    destination_path: str
    file_size: int


class UploadableFile(typing.TypedDict):
    local_path: pathlib.Path
    destination_path: str
    upload_uri: str


class UploadTransaction:
    def __init__(
        self,
        items: collections.abc.Sequence[TransactionFile],
        association: Association,
        device_id: typing.Optional[str] = None,
        origination: typing.Optional[str] = None,
        batch_size: typing.Optional[int] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
        caller_org_id: typing.Optional[str] = None,
    ):
        self.__items = items
        self.__association = association
        self.__device_id = device_id

        if origination is None:
            origination = RobotoEnv.default().roboto_env or f"roboto {roboto_version()}"
        self.__origination = origination

        self.__batch_size = batch_size if batch_size is not None else max(len(items), 1)

        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__caller_org_id = caller_org_id

        self.__transaction_id: typing.Optional[str] = None
        self.__upload_mappings: typing.Optional[dict[str, str]] = None

        self.__completed_upload_node_ids: list[str] = []

        self.__pending_uploads: list[tuple[UploadableFile, FutureLike[None]]] = []

    def __enter__(self) -> UploadTransaction:
        resource_manifest = {file["destination_path"]: file["file_size"] for file in self.__items}
        request = BeginUploadRequest(
            association=self.__association,
            origination=self.__origination,
            resource_manifest=resource_manifest,
            device_id=self.__device_id,
        )

        response = self.__roboto_client.post(
            "v1/files/upload",
            data=request,
            caller_org_id=self.__caller_org_id,
        ).to_record(BeginUploadResponse)

        self.__transaction_id = response.transaction_id
        self.__upload_mappings = response.upload_mappings

        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ):
        if self.__pending_uploads:
            logger.warning(
                f"UploadTransaction exiting with {len(self.__pending_uploads)} "
                "pending uploads. Did you forget to call await_uploads()?"
            )

        if not exc_value:
            # GM(2025-11-18)
            # This matches current Dataset::upload behavior of only completing the transaction if there are no errors.
            self.__finalize()

    def __iter__(self) -> typing.Generator[UploadableFile, None, None]:
        """
        Yields files to upload in batches based on batch_size.

        After each batch is fully yielded, await_uploads() is called automatically
        to wait for all registered uploads in that batch to complete and report
        progress to the API.

        The caller must call register_upload() for each file after initiating
        the S3 transfer. The await_uploads() call at batch boundaries will handle
        waiting for completion and flushing progress.
        """
        for batch_start in range(0, len(self.__items), self.__batch_size):
            batch = self.__items[batch_start : batch_start + self.__batch_size]
            for item in batch:
                yield {
                    "local_path": item["local_path"],
                    "destination_path": item["destination_path"],
                    "upload_uri": self.upload_mappings[item["destination_path"]],
                }
            # After all files in batch are yielded and uploads registered,
            # wait for completion and report progress to API
            self.await_uploads()

    @property
    def completed_upload_node_ids(self) -> list[str]:
        return self.__completed_upload_node_ids

    @property
    def transaction_id(self) -> str:
        if self.__transaction_id is None:
            raise Exception("An UploadTransaction must be used as a context manager.")

        return self.__transaction_id

    @property
    def upload_mappings(self) -> dict[str, str]:
        if self.__upload_mappings is None:
            raise Exception("An UploadTransaction must be used as a context manager.")

        return self.__upload_mappings

    def await_uploads(self) -> None:
        """
        Wait for all registered uploads to complete, then report progress to API.

        Must be called while still inside the object_store context manager,
        since the transfer manager may be shut down when that context exits.

        Raises:
            ExceptionGroup: If any uploads fail, an ExceptionGroup containing all
                           upload errors is raised after reporting successfully
                           completed uploads.
        """
        if not self.__pending_uploads:
            return

        completed: list[UploadableFile] = []
        errors: list[Exception] = []

        for file, future in self.__pending_uploads:
            try:
                future.result()
                completed.append(file)
            except Exception as e:
                logger.error("Upload failed: %s", file["destination_path"], exc_info=e)
                errors.append(e)

        # Report successfully completed uploads to the API
        if completed:
            self.__flush(completed)

        self.__pending_uploads.clear()

        # After reporting successes, raise all errors encountered
        if errors:
            raise ExceptionGroup("One or more uploads failed", errors)

    def make_credential_provider(self) -> CredentialProvider:
        def _get_upload_credentials() -> Credentials:
            response = self.__roboto_client.get(f"v1/files/upload/{self.transaction_id}/credentials").to_record_list(
                RobotoCredentials
            )

            if len(response) == 0:
                raise RobotoInternalException(f"Unable to get upload credentials for transaction {self.transaction_id}")

            creds = response[0]
            return creds.to_object_store_credentials()

        return _get_upload_credentials

    def register_upload(self, file: UploadableFile, future: FutureLike[None]) -> None:
        """
        Register a pending upload future.

        Call this immediately after initiating each S3 transfer.
        The future will be awaited when await_uploads() is called.
        """
        self.__pending_uploads.append((file, future))

    def __finalize(self):
        self.__roboto_client.put(f"v1/files/upload/{self.transaction_id}/complete")

    def __flush(self, batch: list[UploadableFile]):
        manifest_items = [file["upload_uri"] for file in batch]
        response = self.__roboto_client.put(
            f"v1/files/upload/{self.transaction_id}/progress",
            data=ReportUploadProgressRequest(
                manifest_items=manifest_items,
            ),
        )
        node_ids = response.to_string_list()
        self.__completed_upload_node_ids.extend(node_ids)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import pathlib
import tempfile

from cookiecutter.main import cookiecutter

from ..command import RobotoCommand
from ..context import CLIContext

COOKIECUTTER_REPO = "https://github.com/roboto-ai/cookiecutter-roboto-actions.git"


def init(args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser) -> None:
    path: pathlib.Path = args.path
    path = path.resolve()
    print(f"Creating Roboto action package under: {path}")

    if not path.is_dir():
        parser.error(f"{path} does not exist, or is not a directory.")

    with tempfile.TemporaryDirectory() as tmpdir:
        # GM (2025-11-05)
        # cookiecutter currently always includes the following prompt:
        #   > You've downloaded ... before. Is it okay to delete and re-download it? [y/n] (y)
        # I think this is annoying and useless. Also selecting "n" is always the wrong thing to do.
        # cookiecutter maintainers refuse to add an option for always redownloading the template:
        # see https://github.com/cookiecutter/cookiecutter/issues/1201.
        # So, this.
        # Refs:
        #   1. https://github.com/cookiecutter/cookiecutter/blob/main/cookiecutter/main.py#L79-L82
        #   2. https://github.com/cookiecutter/cookiecutter/blob/main/cookiecutter/config.py#L89-L113
        cookiecutter_config = {"cookiecutters_dir": tmpdir}

        try:
            cookiecutter(
                COOKIECUTTER_REPO,
                output_dir=str(path),
                default_config=cookiecutter_config,  # type: ignore
            )
        except KeyboardInterrupt:
            pass


def init_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "path",
        nargs="?",
        type=pathlib.Path,
        help="Existing directory under which an action package will be created.",
        default=pathlib.Path.cwd(),
    )


init_command = RobotoCommand(
    name="init",
    logic=init,
    setup_parser=init_parser,
    command_kwargs={"help": "Initialize a new action package."},
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .clear import clear_command
from .size import size_command
from .where import where_command

commands = [
    clear_command,
    size_command,
    where_command,
]

command_set = RobotoCommandSet(
    name="cache",
    help="Manage the local Roboto cache.",
    commands=commands,
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import multiprocessing
import pathlib
import typing
import urllib.parse

import boto3
import boto3.s3.transfer
import botocore.client
import botocore.config
import botocore.credentials
import botocore.session

from .object_store import (
    CredentialProvider,
    FutureLike,
    ObjectStore,
    OnProgress,
)
from .registry import StoreRegistry


class ProgressCallbackInvoker(boto3.s3.transfer.BaseSubscriber):
    """Invoke a provided callback via a subscriber.

    Taken from https://github.com/boto/boto3/blob/develop/boto3/s3/transfer.py#L505C1-L516C42
    """

    def __init__(self, callback):
        self._callback = callback

    def on_progress(self, bytes_transferred, **kwargs):
        self._callback(bytes_transferred)


@StoreRegistry.register("s3")
class S3Store(ObjectStore):
    @classmethod
    def create(cls, credential_provider: CredentialProvider, **kwargs) -> S3Store:
        """
        Factory function to assemble an S3Store with refreshable credentials.
        """

        # 1. Fetch initial credentials to bootstrap
        creds = credential_provider()

        # 2. Build the RefreshableCredentials object
        refreshable_creds = botocore.credentials.RefreshableCredentials.create_from_metadata(
            metadata=creds,
            refresh_using=credential_provider,
            method="roboto-api",
        )

        # 3. Attach to a Botocore Session
        session = botocore.session.get_session()
        session._credentials = refreshable_creds
        session.set_config_variable("region", creds["region"])

        # 4. Create the high-level Boto3 Session
        boto_session = boto3.Session(botocore_session=session)

        # 5. Create the Client and TransferConfig
        s3_client = boto_session.client("s3", config=botocore.config.Config(tcp_keepalive=True))

        transfer_config = boto3.s3.transfer.TransferConfig(
            use_threads=True, max_concurrency=multiprocessing.cpu_count() * 2
        )

        # 6. Inject/instantiate
        return cls(s3_client, transfer_config=transfer_config)

    def __init__(
        self,
        s3_client: botocore.client.BaseClient,
        transfer_config: typing.Optional[boto3.s3.transfer.TransferConfig] = None,
    ):
        config = transfer_config or boto3.s3.transfer.TransferConfig()
        self.__transfer_manager = boto3.s3.transfer.create_transfer_manager(s3_client, config)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__transfer_manager.shutdown()

    def put(
        self, source: pathlib.Path, destination_uri: str, on_progress: typing.Optional[OnProgress] = None
    ) -> FutureLike[None]:
        parsed_uri = urllib.parse.urlparse(destination_uri)
        bucket = parsed_uri.netloc
        key = parsed_uri.path.lstrip("/")

        subscribers = []
        if on_progress is not None:
            subscribers.append(ProgressCallbackInvoker(on_progress))

        return self.__transfer_manager.upload(
            str(source),
            bucket,
            key,
            subscribers=subscribers,
        )

    def get(
        self, source_uri: str, destination: pathlib.Path, on_progress: typing.Optional[OnProgress] = None
    ) -> FutureLike[None]:
        destination.parent.mkdir(parents=True, exist_ok=True)

        parsed_uri = urllib.parse.urlparse(source_uri)
        bucket = parsed_uri.netloc
        key = parsed_uri.path.lstrip("/")

        subscribers = []
        if on_progress is not None:
            subscribers.append(ProgressCallbackInvoker(on_progress))

        return self.__transfer_manager.download(
            bucket,
            key,
            str(destination),
            subscribers=subscribers,
        )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import pathlib
import sys
import typing

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

from ..association import Association
from ..auth import Permissions
from ..exceptions import RobotoInternalException, RobotoNotFoundException
from ..http import RobotoClient
from ..logging import default_logger
from .credentials import RobotoCredentials
from .object_store import (
    CredentialProvider,
    Credentials,
    FutureLike,
)

logger = default_logger()


class DownloadableFile(typing.TypedDict):
    """A file to be downloaded from the Roboto Platform."""

    bucket_name: str
    """Name of the bucket where the file is stored."""

    source_uri: str
    """Full URI of the file in cloud storage (e.g., 's3://bucket/key')."""

    destination_path: pathlib.Path
    """Local path where the file should be saved."""


class DownloadSession:
    """Manages a batch of file downloads with shared credentials.

    Provides credential lifecycle management and async operation coordination
    for downloading files from a dataset. Unlike UploadTransaction, this does
    not manage an API transaction since downloads don't require server-side
    state tracking.

    Example:
        >>> session = DownloadSession(
        ...     items=[
        ...         {"bucket_name": "bucket", "source_uri": "s3://bucket/key", "destination_path": Path("/tmp/file")}
        ...     ],
        ...     association=Association.dataset("ds_123"),
        ... )
        >>> object_store = registry.get_store_for_uri(uri, session.credential_provider)
        >>> with object_store:
        ...     for file in session:
        ...         future = object_store.get(file["source_uri"], file["destination_path"])
        ...         session.register_download(file, future)
    """

    def __init__(
        self,
        items: collections.abc.Sequence[DownloadableFile],
        association: Association,
        roboto_client: typing.Optional[RobotoClient] = None,
        caller_org_id: typing.Optional[str] = None,
    ):
        """Initialize a download session.

        Args:
            items: Sequence of files to download.
            association: Association of the files to download.
            roboto_client: Optional Roboto client for API calls.
            caller_org_id: Optional organization ID for cross-org access.
        """
        self.__items = items
        self.__association = association
        if not self.__association.is_dataset:
            raise ValueError("Roboto currently only supports downloading files from datasets.")

        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__caller_org_id = caller_org_id

        self.__pending_downloads: list[tuple[DownloadableFile, FutureLike[None]]] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.await_downloads()

    def __iter__(self) -> typing.Generator[DownloadableFile, None, None]:
        """Yield files to download."""
        for item in self.__items:
            yield item

        self.await_downloads()

    def make_credential_provider(self, bucket_name: typing.Optional[str] = None) -> CredentialProvider:
        """Return a credential provider for read-only access to the dataset."""

        def _get_download_credentials() -> Credentials:
            query_params: dict[str, str] = {"mode": Permissions.ReadOnly.value}
            response = self.__roboto_client.get(
                f"v1/datasets/id/{self.__dataset_id}/credentials",
                query=query_params,
                caller_org_id=self.__caller_org_id,
            ).to_record_list(RobotoCredentials)

            if len(response) == 0:
                raise RobotoInternalException(f"Unable to get credentials for {self.__association!r}")

            if bucket_name is None:
                return response[0].to_object_store_credentials()
            else:
                for creds in response:
                    if creds.bucket == bucket_name:
                        return creds.to_object_store_credentials()

                raise RobotoNotFoundException(
                    f"Unable to get credentials for bucket {bucket_name}. "
                    "Is it properly registered with the Roboto platform?"
                )

        return _get_download_credentials

    @property
    def __dataset_id(self) -> str:
        dataset_id = self.__association.dataset_id
        if dataset_id is None:
            raise ValueError("Roboto currently only supports downloading files from datasets.")
        return dataset_id

    def await_downloads(self) -> None:
        """Wait for all registered downloads to complete.

        Must be called while still inside the object_store context manager,
        since the transfer manager may be shut down when that context exits.

        Raises:
            ExceptionGroup: If any downloads fail, an ExceptionGroup containing all
                           download errors is raised.
        """
        if not self.__pending_downloads:
            return

        errors: list[Exception] = []

        for file, future in self.__pending_downloads:
            try:
                future.result()
            except Exception as e:
                logger.error("Download failed: %s", file["source_uri"], exc_info=e)
                errors.append(e)

        self.__pending_downloads.clear()

        if errors:
            raise ExceptionGroup("One or more downloads failed", errors)

    def register_download(self, file: DownloadableFile, future: FutureLike[None]) -> None:
        """Register a pending download future.

        Call this immediately after initiating each download transfer.
        The future will be awaited when await_downloads() is called.
        """
        self.__pending_downloads.append((file, future))



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections
import collections.abc
import datetime
import decimal
import json
import re
import typing

import pydantic
import pydantic_core

from ..collection_utils import get_by_path
from ..compat import StrEnum

ConditionValue = typing.Optional[typing.Union[str, bool, int, float, decimal.Decimal, datetime.datetime]]


class Comparator(StrEnum):
    """The comparator to use when comparing a field to a value."""

    Equals = "EQUALS"
    NotEquals = "NOT_EQUALS"
    GreaterThan = "GREATER_THAN"
    GreaterThanOrEqual = "GREATER_THAN_OR_EQUAL"
    LessThan = "LESS_THAN"
    LessThanOrEqual = "LESS_THAN_OR_EQUAL"
    Contains = "CONTAINS"
    NotContains = "NOT_CONTAINS"
    IsNull = "IS_NULL"
    IsNotNull = "IS_NOT_NULL"
    Exists = "EXISTS"
    NotExists = "NOT_EXISTS"
    BeginsWith = "BEGINS_WITH"
    Like = "LIKE"  # SQL Syntax
    NotLike = "NOT_LIKE"  # SQL Syntax

    def to_compact_string(self):
        if self is Comparator.Equals:
            return "="
        elif self is Comparator.NotEquals:
            return "!="
        elif self is Comparator.GreaterThanOrEqual:
            return ">"
        elif self is Comparator.GreaterThan:
            return ">="
        elif self is Comparator.LessThan:
            return "<"
        elif self is Comparator.LessThanOrEqual:
            return "<="
        else:
            return self.value

    @staticmethod
    def from_string(value: str) -> "Comparator":
        as_comparator: typing.Optional[Comparator] = {
            "EQUALS": Comparator.Equals,
            "=": Comparator.Equals,
            "NOT_EQUALS": Comparator.NotEquals,
            "!=": Comparator.NotEquals,
            "GREATER_THAN": Comparator.GreaterThan,
            ">": Comparator.GreaterThan,
            "GREATER_THAN_OR_EQUAL": Comparator.GreaterThanOrEqual,
            ">=": Comparator.GreaterThanOrEqual,
            "LESS_THAN": Comparator.LessThan,
            "<": Comparator.LessThan,
            "LESS_THAN_OR_EQUAL": Comparator.LessThanOrEqual,
            "<=": Comparator.LessThanOrEqual,
            "CONTAINS": Comparator.Contains,
            "NOT_CONTAINS": Comparator.NotContains,
            "IS_NULL": Comparator.IsNull,
            "IS_NOT_NULL": Comparator.IsNotNull,
            "EXISTS": Comparator.Exists,
            "NOT_EXISTS": Comparator.NotExists,
            "BEGINS_WITH": Comparator.BeginsWith,
            "LIKE": Comparator.Like,
            "NOT_LIKE": Comparator.NotLike,
        }.get(value.upper())

        if as_comparator is None:
            raise ValueError(f"Unrecognized comparator {value}")

        return as_comparator


class FieldTarget(typing.NamedTuple):
    """An immutable field path with an optional resource type hint for query interpretation.

    FieldTarget pairs a field path with an optional resource qualifier to help the query
    system resolve which resource type the field belongs to (dataset, file, topic, or message_path).
    When the resource is omitted, the query context infers it.

    Examples:
        >>> target = FieldTarget("org_id", "dataset")
        >>> target.path
        'org_id'
        >>> target.resource
        'dataset'

        >>> target = FieldTarget("metadata.min", None)
        >>> target.resource is None
        True
    """

    path: str
    """Field path within the target resource, using dot notation for nested fields."""

    resource: typing.Optional[typing.Literal["dataset", "file", "topic", "message_path"]]
    """Optional resource type hint to disambiguate field path interpretation."""


class Field(str):
    """A string-like field path that parses resource qualifiers and extracts the target path.

    Field extends str to provide automatic parsing of qualified field paths like
    "dataset.metadata.owner" or "topic.name" into their constituent parts. It identifies
    the target resource type (dataset, file, topic, or message_path) and extracts the
    actual field path within that resource.

    The class supports both qualified paths (e.g., "dataset.org_id") and unqualified paths
    (e.g., "org_id"). For qualified paths, it strips the resource prefix and stores both
    the original fully qualified path and the extracted target information.

    Examples:
        >>> field = Field("dataset.metadata.foo")
        >>> field.target.resource
        'dataset'
        >>> field.target.path
        'metadata.foo'

        >>> field = Field("org_id")
        >>> field.target.resource is None
        True
        >>> field.target.path
        'org_id'
    """

    __MESSAGE_PATH_REGEX: typing.ClassVar[re.Pattern] = re.compile(r"msgpath[.](?P<msgpath_property>.+)")
    __TOPIC_REGEX: typing.ClassVar[re.Pattern] = re.compile(r"(?:msgpath[.])?topic[.](?P<topic_property>.+)")
    __FILE_REGEX: typing.ClassVar[re.Pattern] = re.compile(r"(?:msgpath[.])?(?:topic[.])?file[.](?P<file_property>.+)")
    __DATASET_REGEX: typing.ClassVar[re.Pattern] = re.compile(
        r"(?:msgpath[.])?(?:topic[.])?(?:file[.])?dataset[.](?P<dataset_property>.+)"
    )

    __fully_qualified_path: str
    __target_path: str
    __target_resource: typing.Optional[typing.Literal["dataset", "file", "topic", "message_path"]] = None

    @classmethod
    def wrap(cls, field: typing.Union[str, "Field"]) -> Field:
        if isinstance(field, Field):
            return field
        else:
            return cls(field)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: typing.Any, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        """Define the Pydantic core schema for validation.

        This follows the pattern from _SecretField in pydantic/types.py:
        (https://github.com/pydantic/pydantic/blob/f42171c760d43b9522fde513ae6e209790f7fefb/pydantic/types.py#L1746C7-L1746C19)
        """

        def get_json_schema(
            _core_schema: pydantic_core.CoreSchema,
            handler: pydantic.GetJsonSchemaHandler,
        ) -> pydantic.json_schema.JsonSchemaValue:
            return handler(pydantic_core.core_schema.str_schema())

        def get_schema(strict: bool) -> pydantic_core.CoreSchema:
            # Create inner schema with strict mode
            inner_schema = {**pydantic_core.core_schema.str_schema(), "strict": strict}

            # JSON schema: validate string, then construct Field instance
            json_schema = pydantic_core.core_schema.no_info_after_validator_function(
                cls,
                inner_schema,  # type: ignore[arg-type]
            )

            return pydantic_core.core_schema.json_or_python_schema(
                python_schema=pydantic_core.core_schema.union_schema(
                    [
                        # Accept existing Field instances
                        pydantic_core.core_schema.is_instance_schema(cls),
                        # Or validate and construct from string
                        json_schema,
                    ],
                    custom_error_type="string_type",
                ),
                json_schema=json_schema,
                serialization=pydantic_core.core_schema.to_string_ser_schema(),
            )

        # Return lax or strict schema based on mode
        return pydantic_core.core_schema.lax_or_strict_schema(
            lax_schema=get_schema(strict=False),
            strict_schema=get_schema(strict=True),
            metadata={"pydantic_js_functions": [get_json_schema]},
        )

    def __init__(self, path: str):
        self.__fully_qualified_path = path
        self.__parse()

    @property
    def target(self) -> FieldTarget:
        return FieldTarget(path=self.__target_path, resource=self.__target_resource)

    def __parse(self) -> None:
        dataset_match = self.__DATASET_REGEX.match(self.__fully_qualified_path)
        if dataset_match:
            self.__target_resource = "dataset"
            self.__target_path = dataset_match.group("dataset_property")
            return

        file_match = self.__FILE_REGEX.match(self.__fully_qualified_path)
        if file_match:
            self.__target_resource = "file"
            self.__target_path = file_match.group("file_property")
            return

        topic_match = self.__TOPIC_REGEX.match(self.__fully_qualified_path)
        if topic_match:
            self.__target_resource = "topic"
            self.__target_path = topic_match.group("topic_property")
            return

        msgpath_match = self.__MESSAGE_PATH_REGEX.match(self.__fully_qualified_path)
        if msgpath_match:
            self.__target_resource = "message_path"
            self.__target_path = msgpath_match.group("msgpath_property")
            return

        self.__target_path = self.__fully_qualified_path


class Condition(pydantic.BaseModel):
    """A filter for any arbitrary attribute for a Roboto resource."""

    field: typing.Annotated[str, Field]
    comparator: Comparator
    value: ConditionValue = None

    @classmethod
    def equals_cond(cls, field: str, value: ConditionValue) -> "Condition":
        return cls(field=field, comparator=Comparator.Equals, value=value)

    def __str__(self):
        base_condition = f"{self.field} {self.comparator.to_compact_string()}"
        if self.value is None:
            return base_condition

        if isinstance(self.value, datetime.datetime):
            return f"{base_condition} {self.value.isoformat()}"

        return f"{base_condition} {json.dumps(self.value)}"

    def __repr__(self):
        return self.model_dump_json()

    def matches(self, target: dict) -> bool:
        field = Field.wrap(self.field)
        value = get_by_path(target, field.target.path.split("."))

        if self.comparator in [Comparator.NotExists, Comparator.IsNull]:
            return value is None

        if self.comparator in [Comparator.Exists, Comparator.IsNotNull]:
            return value is not None

        # We need the value for everything else
        if value is None:
            return False

        if isinstance(value, str) and not isinstance(self.value, str):
            if isinstance(self.value, int):
                value = int(value)
            elif isinstance(self.value, float):
                value = float(value)
            elif isinstance(self.value, bool):
                value = value.lower() == "true"
            elif isinstance(self.value, decimal.Decimal):
                value = decimal.Decimal.from_float(float(value))
            elif isinstance(self.value, datetime.datetime):
                value = datetime.datetime.fromisoformat(value)

        if self.comparator is Comparator.Equals:
            return value == self.value

        if self.comparator is Comparator.NotEquals:
            return value != self.value

        if self.comparator is Comparator.GreaterThan:
            return value > self.value

        if self.comparator is Comparator.GreaterThanOrEqual:
            return value >= self.value

        if self.comparator is Comparator.LessThan:
            return value < self.value

        if self.comparator is Comparator.LessThanOrEqual:
            return value <= self.value

        if self.comparator is Comparator.Contains:
            return isinstance(value, collections.abc.Collection) and self.value in value

        if self.comparator is Comparator.NotContains:
            return isinstance(value, collections.abc.Collection) and self.value not in value

        if self.comparator is Comparator.BeginsWith:
            if not isinstance(value, str) or not isinstance(self.value, str):
                return False

            return value.startswith(self.value)

        return False

    def target_unspecified(self) -> bool:
        """Is the target resource of this query condition unspecified?"""
        return Field.wrap(self.field).target.resource is None

    def targets_dataset(self) -> bool:
        """Does this query condition target a dataset?"""
        return Field.wrap(self.field).target.resource == "dataset"

    def targets_file(self) -> bool:
        """Does this query condition target a file?"""
        return Field.wrap(self.field).target.resource == "file"

    def targets_topic(self) -> bool:
        """Does this query condition target a topic?"""
        return Field.wrap(self.field).target.resource == "topic"

    def targets_message_path(self) -> bool:
        """Does this query condition target a message path?"""
        return Field.wrap(self.field).target.resource == "message_path"


class ConditionOperator(StrEnum):
    """The operator to use when combining multiple conditions."""

    And = "AND"
    Or = "OR"
    Not = "NOT"

    @staticmethod
    def from_string(value: str) -> "ConditionOperator":
        if value.lower() == "and":
            return ConditionOperator.And
        elif value.lower() == "or":
            return ConditionOperator.Or
        elif value.lower() == "not":
            return ConditionOperator.Not

        raise ValueError(f"Unrecognized Operator {value}")


class ConditionGroup(pydantic.BaseModel):
    """A group of conditions that are combined together."""

    operator: ConditionOperator
    conditions: collections.abc.Sequence[ConditionType]

    def matches(self, target: dict | typing.Callable[[ConditionType], bool]):
        if callable(target):
            conditions_match = [target(condition) for condition in self.conditions]
        else:
            conditions_match = [condition.matches(target) for condition in self.conditions]

        if self.operator is ConditionOperator.And:
            return all(conditions_match)

        if self.operator is ConditionOperator.Or:
            return any(conditions_match)

        if self.operator is ConditionOperator.Not:
            return not any(conditions_match)

        return False

    @staticmethod
    def or_group(*conditions: ConditionType) -> "ConditionGroup":
        return ConditionGroup(operator=ConditionOperator.Or, conditions=conditions)

    @staticmethod
    def and_group(*conditions: ConditionType) -> "ConditionGroup":
        return ConditionGroup(operator=ConditionOperator.And, conditions=conditions)

    @pydantic.field_validator("conditions")
    def validate_conditions(cls, v: collections.abc.Sequence[ConditionType]):
        if len(v) == 0:
            raise ValueError("At least one condition must be provided to a ConditionGroup, got 0!")

        return v

    def __str__(self):
        return "(" + f" {self.operator.value} ".join([str(condition) for condition in self.conditions]) + ")"

    def __repr__(self):
        return self.model_dump_json()


ConditionType = typing.Union[Condition, ConditionGroup]



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic


class ChatStartTextEvent(pydantic.BaseModel):
    """Signals the beginning of text generation in a chat response."""


class ChatTextDeltaEvent(pydantic.BaseModel):
    """Contains incremental text content as the AI generates its response."""

    text: str
    """Text fragment from the streaming response."""


class ChatTextEndEvent(pydantic.BaseModel):
    """Signals the completion of text generation in a chat response."""


class ChatToolUseEvent(pydantic.BaseModel):
    """Signals that the AI is invoking a tool to gather information."""

    name: str
    """Name of the tool being invoked."""

    tool_use_id: str
    """Unique identifier for this tool invocation."""


class ChatToolResultEvent(pydantic.BaseModel):
    """Contains the result of a tool invocation."""

    name: str
    """Name of the tool that was invoked."""

    tool_use_id: str
    """Unique identifier for this tool invocation."""

    success: bool
    """Whether the tool invocation succeeded."""


ChatEvent: typing.TypeAlias = typing.Union[
    ChatStartTextEvent,
    ChatTextDeltaEvent,
    ChatTextEndEvent,
    ChatToolUseEvent,
    ChatToolResultEvent,
]



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import typing
import urllib.parse

from .object_store import (
    CredentialProvider,
    ObjectStore,
)


class StoreRegistry:
    _registry: typing.ClassVar[dict[str, typing.Type[ObjectStore]]] = {}

    @classmethod
    def register(cls, scheme: str):
        """
        Class Decorator.
        Registers the decorated class to handle the specific URI scheme.
        """

        def wrapper(store_cls: typing.Type[ObjectStore]):
            if scheme in cls._registry:
                raise ValueError(f"Scheme '{scheme}' is already registered to {cls._registry[scheme]}")
            cls._registry[scheme] = store_cls
            return store_cls

        return wrapper

    @classmethod
    def get_store_for_uri(cls, uri: str, credential_provider: CredentialProvider, **kwargs) -> ObjectStore:
        parsed = urllib.parse.urlparse(uri)
        scheme = parsed.scheme
        if scheme not in cls._registry:
            raise ValueError(f"No ObjectStore class registered for scheme: '{scheme}://'")

        store_class = cls._registry[scheme]

        # Delegate construction to the specific class
        return store_class.create(credential_provider=credential_provider, **kwargs)



# Copyright (c) 2026 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic

from ..association import Association


class AbortTransactionsRequest(pydantic.BaseModel):
    """Request payload for aborting file upload transactions.

    Used to cancel ongoing file upload transactions, typically when uploads
    fail or are no longer needed. This cleans up any reserved resources
    and marks associated files as no longer pending.
    """

    transaction_ids: list[str]
    """List of transaction IDs to abort."""


class BeginUploadRequest(pydantic.BaseModel):
    """Request payload to begin a batch file upload transaction.

    Used to initiate a multi-file upload transaction for any association type
    (dataset, topic, etc.). Returns a transaction ID and upload mappings that
    specify where each file should be uploaded.
    """

    association: Association
    """The entity these files will be associated with (e.g., dataset, topic)."""

    origination: str
    """Description of the upload source (e.g., 'roboto-sdk v1.0.0')."""

    resource_manifest: dict[str, int]
    """Dictionary mapping destination file paths to file sizes in bytes."""

    device_id: typing.Optional[str] = None
    """Optional identifier of the device that generated this data."""


class BeginUploadResponse(pydantic.BaseModel):
    """Response from beginning a batch upload transaction.

    Contains the transaction ID needed for subsequent progress reporting
    and completion calls, plus mappings from file paths to their upload URIs.
    """

    transaction_id: str
    """Unique identifier for this upload transaction."""

    upload_mappings: dict[str, str]
    """Dictionary mapping file paths to their S3 upload URIs."""


class BeginSignedUrlUploadRequest(pydantic.BaseModel):
    """Request payload to begin a single file upload with a signed URL.

    Used for simpler upload scenarios where a pre-signed URL is preferred
    over temporary credentials. The returned URL can be used directly for
    uploading the file content.
    """

    association: Association
    """The entity this file will be associated with (e.g., dataset, topic)."""

    file_path: str
    """Destination path for the file within the association."""

    file_size: int
    """Size of the file in bytes."""

    origination: typing.Optional[str] = None
    """Optional description of the upload source."""


class BeginSignedUrlUploadResponse(pydantic.BaseModel):
    """Response from beginning a single file upload.

    Contains the upload ID for completing the transaction and a pre-signed
    URL that can be used to upload the file content directly.
    """

    upload_id: str
    """Unique identifier for this upload transaction."""

    upload_url: str
    """Pre-signed URL for uploading the file content."""


class ReportUploadProgressRequest(pydantic.BaseModel):
    """Request payload for reporting file upload progress.

    Used to notify the platform about the completion status of individual
    files within a batch upload transaction. This enables progress tracking
    and partial completion handling for large file uploads.
    """

    manifest_items: list[str]
    """List of file URIs that have completed upload."""



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .object_store import (
    CredentialProvider,
    Credentials,
    FutureLike,
    ObjectStore,
    OnProgress,
)
from .registry import StoreRegistry
from .s3 import S3Store

__all__ = (
    "Credentials",
    "CredentialProvider",
    "FutureLike",
    "ObjectStore",
    "OnProgress",
    "StoreRegistry",
    "S3Store",
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import pathlib
import re
import typing

from ....association import Association
from ....exceptions import IngestionException
from ....fs import FileService
from ....http import RobotoClient
from ....logging import default_logger

logger = default_logger()


def upload_representation_file(
    file_path: pathlib.Path,
    association: Association,
    caller_org_id: typing.Optional[str] = None,
    roboto_client: typing.Optional[RobotoClient] = None,
) -> str:
    dest_path = pathlib.Path(".VISUALIZATION_ASSETS") / file_path.name
    file_service = FileService(roboto_client)
    file_ids = file_service.upload(
        [file_path],
        association,
        destination_paths={file_path: str(dest_path)},
        caller_org_id=caller_org_id,
    )

    if not file_ids:
        raise IngestionException(
            f"Failed to upload visualization asset to {association.association_id}. "
            "If this error persists after retry, please reach out to Roboto support."
        )

    return file_ids[0]


def make_topic_filename_safe(name: str, replacement_char: str = "_") -> str:
    unsafe_chars = re.compile(r'[<>:"/\\|?*\0]')
    return unsafe_chars.sub(replacement_char, name)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import pathlib
import typing
import urllib.request

from ....association import AssociationType
from ....compat import import_optional_dependency
from ....http import RobotoClient
from ....logging import default_logger
from ..operations import (
    MessagePathRepresentationMapping,
)
from ..record import (
    CanonicalDataType,
    MessagePathRecord,
    RepresentationRecord,
    RepresentationStorageFormat,
)
from ..topic_reader import Timestamp, TopicReader
from .table_transforms import (
    compute_time_filter_mask,
    extract_timestamp_field,
    extract_timestamps,
    resolve_columns,
    should_read_row_group,
)
from .timestamp import Timestamp as TimestampDescriptor

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep
    import pyarrow  # pants: no-infer-dep
    import pyarrow.parquet  # pants: no-infer-dep


logger = default_logger()

OUTFILE_NAME_PATTERN = "{repr_id}_{file_id}.parquet"
"""Filename template for locally cached Parquet files."""

_COLUMN_COUNT_LOCAL_CACHE_THRESHOLD = 10
"""
When the number of columns to read meets or exceeds this threshold,
the Parquet file is downloaded to a local cache before reading.
Reading many columns over HTTP incurs significant per-column overhead;
a local file avoids repeated network round-trips, but also loads much unnecessary data.
10 was chosen as individual column loading can be very slow for some files and many columns, while local
cache based access is never terribly slow.
"""


class _ReadContext(typing.NamedTuple):
    """Shared state prepared for reading a Parquet-backed topic."""

    parquet_file: pyarrow.parquet.ParquetFile
    timestamp_field: TimestampDescriptor
    columns: list[str]
    include_timestamp_column: bool


class ParquetTopicReader(TopicReader):
    """Private interface for retrieving topic data stored in Parquet files.

    Note:
        This is not intended as a public API.
        To access topic data, prefer the ``get_data`` or ``get_data_as_df`` methods
        on :py:class:`~roboto.domain.topics.Topic`, :py:class:`~roboto.domain.topics.MessagePath`,
        or :py:class:`~roboto.domain.events.Event`.
    """

    __cache_dir: typing.Optional[pathlib.Path]
    __roboto_client: RobotoClient

    @staticmethod
    def accepts(
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
    ) -> bool:
        for mapping in message_paths_to_representations:
            if mapping.representation.storage_format != RepresentationStorageFormat.PARQUET:
                return False
        return True

    def __init__(
        self,
        roboto_client: RobotoClient,
        cache_dir: typing.Optional[pathlib.Path] = None,
    ):
        self.__roboto_client = roboto_client
        self.__cache_dir = cache_dir

    def get_data(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
        start_time: typing.Optional[int] = None,
        end_time: typing.Optional[int] = None,
        timestamp_message_path_representation_mapping: typing.Optional[MessagePathRepresentationMapping] = None,
    ) -> collections.abc.Generator[tuple[Timestamp, dict[str, typing.Any]], None, None]:
        pc = import_optional_dependency("pyarrow.compute", "analytics")

        if timestamp_message_path_representation_mapping is None:
            raise NotImplementedError(
                "Reading data from a Parquet file requires one column to be marked as a 'CanonicalDataType.Timestamp'. "
                "This is likely an issue with data ingestion. Please reach out to Roboto support."
            )

        ctx = self.__prepare_read(
            message_paths_to_representations,
            timestamp_message_path_representation_mapping,
        )
        if ctx is None:
            return

        for row_group_idx in range(ctx.parquet_file.metadata.num_row_groups):
            row_group_metadata = ctx.parquet_file.metadata.row_group(row_group_idx)
            if not should_read_row_group(
                row_group_metadata,
                ctx.timestamp_field,
                start_time,
                end_time,
            ):
                continue

            row_group_table = ctx.parquet_file.read_row_group(
                row_group_idx,
                columns=ctx.columns,
            )

            timestamps = extract_timestamps(row_group_table, ctx.timestamp_field)

            # Drop the timestamp column if it wasn't originally requested
            if not ctx.include_timestamp_column:
                row_group_table = row_group_table.drop_columns(ctx.timestamp_field.field.name)

            filter_mask = compute_time_filter_mask(timestamps, start_time, end_time)
            if filter_mask is not None:
                row_group_table = pc.filter(row_group_table, filter_mask)
                timestamps = pc.filter(timestamps, filter_mask)

            # Yield tuples of (timestamp, row_dict)
            for idx, row in enumerate(row_group_table.to_pylist()):
                timestamp = timestamps[idx]
                if pc.is_null(timestamp, nan_is_null=True):
                    logger.warning("Skipping row %d, timestamp is null", idx)
                    continue

                yield timestamp.as_py(), row

    def get_data_as_df(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
        start_time: typing.Optional[int] = None,
        end_time: typing.Optional[int] = None,
        timestamp_message_path_representation_mapping: typing.Optional[MessagePathRepresentationMapping] = None,
    ) -> tuple[pandas.Series, pandas.DataFrame]:
        pd = import_optional_dependency("pandas", "analytics")
        pa = import_optional_dependency("pyarrow", "analytics")
        pc = import_optional_dependency("pyarrow.compute", "analytics")

        if timestamp_message_path_representation_mapping is None:
            raise NotImplementedError(
                "Reading data from a Parquet file requires one column to be marked as a 'CanonicalDataType.Timestamp'. "
                "This is likely an issue with data ingestion. Please reach out to Roboto support."
            )

        ctx = self.__prepare_read(
            message_paths_to_representations,
            timestamp_message_path_representation_mapping,
        )
        if ctx is None:
            return pd.Series(), pd.DataFrame()

        timestamps = []
        tables = []

        for row_group_idx in range(ctx.parquet_file.metadata.num_row_groups):
            row_group_metadata = ctx.parquet_file.metadata.row_group(row_group_idx)
            if not should_read_row_group(
                row_group_metadata,
                ctx.timestamp_field,
                start_time,
                end_time,
            ):
                continue

            row_group_table = ctx.parquet_file.read_row_group(
                row_group_idx,
                columns=ctx.columns,
            )

            row_group_timestamps = extract_timestamps(row_group_table, ctx.timestamp_field)

            if not ctx.include_timestamp_column:
                # The timestamp column was not included in the column projection list.
                row_group_table = row_group_table.drop_columns(
                    ctx.timestamp_field.field.name,
                )

            filter_mask = compute_time_filter_mask(row_group_timestamps, start_time, end_time)
            if filter_mask is not None:
                row_group_table = pc.filter(row_group_table, filter_mask)
                row_group_timestamps = pc.filter(row_group_timestamps, filter_mask)

            timestamps.append(row_group_timestamps)
            tables.append(row_group_table)

        if not tables:
            return pd.Series(), pd.DataFrame()

        combined_timestamps: pyarrow.Array = pa.concat_arrays(timestamps)
        combined_tables: pyarrow.Table = pa.concat_tables(tables)

        return combined_timestamps.to_pandas(), combined_tables.to_pandas()

    def __ensure_single_parquet_file_per_topic(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
    ):
        """
        Support pulling data out of a single Parquet file per Topic.
        This is a non-essential limitation; it is done for expediency of initial implementation.
        This class can and should be extended to support splitting a Topic's MessagePaths
        across multiple underlying files.
        """
        repr_ids = {mapping.representation.representation_id for mapping in message_paths_to_representations}
        if len(repr_ids) > 1:
            raise NotImplementedError(
                "Support for reading data for topics whose data is split across multiple Parquet files  "
                "is not yet implemented. "
                "This is likely an issue with data ingestion. Please reach out to Roboto support."
            )

    def __get_signed_url_for_representation_file(self, representation: RepresentationRecord) -> str:
        association = representation.association
        if association.association_type != AssociationType.File:
            raise NotImplementedError(
                "Unable to get topic data. "
                "Expected the data to be stored in a Parquet file, "
                f"but received a pointer to a '{association.association_type.value}' instead. "
                "This is likely a problem with data ingestion. Please reach out to Roboto support."
            )
        file_id = representation.association.association_id
        logger.debug("Getting signed url for file '%s'", file_id)
        signed_url_response = self.__roboto_client.get(f"v1/files/{file_id}/signed-url")
        return signed_url_response.to_dict(json_path=["data", "url"])

    def __parquet_file_from_remote_streaming(self, representation: RepresentationRecord) -> pyarrow.parquet.ParquetFile:
        """Open a Parquet file over HTTP via a signed URL (no local download)."""
        fs = import_optional_dependency("pyarrow.fs", "analytics")
        fsspec_http = import_optional_dependency("fsspec.implementations.http", "analytics")
        pq = import_optional_dependency("pyarrow.parquet", "analytics")

        http_fs = fsspec_http.HTTPFileSystem()
        signed_url = self.__get_signed_url_for_representation_file(representation)
        return pq.ParquetFile(signed_url, filesystem=fs.PyFileSystem(fs.FSSpecHandler(http_fs)))

    def __should_use_local_cache(self, column_count: int) -> bool:
        """Decide whether the Parquet file should be downloaded to local cache.

        Returns ``True`` when the column count meets or exceeds
        ``_COLUMN_COUNT_LOCAL_CACHE_THRESHOLD`` **and** a cache directory has
        been configured.
        """
        return self.__cache_dir is not None and column_count >= _COLUMN_COUNT_LOCAL_CACHE_THRESHOLD

    def __parquet_file_from_local_cache(self, representation: RepresentationRecord) -> pyarrow.parquet.ParquetFile:
        """Download the Parquet file to local cache (if not already cached) and open it.

        Follows the same caching pattern used by :class:`McapTopicReader`:
        files are stored under ``<cache_dir>/<repr_id>_<file_id>.parquet`` and
        re-used across calls.
        """
        pq = import_optional_dependency("pyarrow.parquet", "analytics")

        assert self.__cache_dir is not None  # guaranteed by __should_use_local_cache

        outfile = self.__cache_dir / OUTFILE_NAME_PATTERN.format(
            repr_id=representation.representation_id,
            file_id=representation.association.association_id,
        )

        if not outfile.exists():
            signed_url = self.__get_signed_url_for_representation_file(representation)
            logger.debug(
                "Downloading Parquet file for representation '%s' to %s",
                representation.representation_id,
                outfile,
            )
            urllib.request.urlretrieve(signed_url, str(outfile))

        return pq.ParquetFile(outfile)

    def __prepare_read(
        self,
        message_paths_to_representations: collections.abc.Iterable[MessagePathRepresentationMapping],
        timestamp_message_path_representation_mapping: MessagePathRepresentationMapping,
    ) -> typing.Optional[_ReadContext]:
        """Shared preamble for ``get_data`` and ``get_data_as_df``.

        Validates inputs, opens the Parquet file (remotely or from local cache),
        resolves the timestamp field, and builds the column projection list.

        When the number of requested columns meets or exceeds
        ``_COLUMN_COUNT_LOCAL_CACHE_THRESHOLD`` and a *cache_dir* was provided,
        the file is downloaded to local disk first to avoid excessive HTTP
        round-trips.

        Returns ``None`` when *message_paths_to_representations* is empty
        (i.e. there is nothing to read).
        """
        self.__ensure_single_parquet_file_per_topic(
            [
                *message_paths_to_representations,
                timestamp_message_path_representation_mapping,
            ]
        )

        mapping = next(iter(message_paths_to_representations), None)
        if mapping is None:
            return None

        timestamp_message_path = self.__timestamp_message_path(timestamp_message_path_representation_mapping)

        # Use message-path count as a proxy for column count to decide
        # whether to download the file locally before opening it.
        estimated_column_count = len(list(mapping.message_paths))
        if self.__should_use_local_cache(estimated_column_count):
            logger.debug(
                "Using local cache for Parquet file (%d message paths >= %d threshold)",
                estimated_column_count,
                _COLUMN_COUNT_LOCAL_CACHE_THRESHOLD,
            )
            parquet_file = self.__parquet_file_from_local_cache(mapping.representation)
        else:
            logger.debug(
                "Using remote HTTP for Parquet file (%d message paths < %d threshold)",
                estimated_column_count,
                _COLUMN_COUNT_LOCAL_CACHE_THRESHOLD,
            )
            parquet_file = self.__parquet_file_from_remote_streaming(mapping.representation)

        columns = resolve_columns(parquet_file.schema_arrow, mapping.message_paths)

        # Even if the timestamp column wasn't requested in the column projection list,
        # request the data to enable timestamp filtering
        include_timestamp_column = timestamp_message_path.source_path in columns
        if not include_timestamp_column:
            columns.append(timestamp_message_path.source_path)

        timestamp_field = extract_timestamp_field(parquet_file.schema_arrow, timestamp_message_path)

        return _ReadContext(
            parquet_file=parquet_file,
            timestamp_field=timestamp_field,
            columns=columns,
            include_timestamp_column=include_timestamp_column,
        )

    def __timestamp_message_path(
        self, message_path_representation_mapping: MessagePathRepresentationMapping
    ) -> MessagePathRecord:
        for message_path in message_path_representation_mapping.message_paths:
            if message_path.canonical_data_type != CanonicalDataType.Timestamp:
                continue

            return message_path

        raise Exception(
            "Could not determine timestamp for topic ingested as Parquet. "
            "This is likely a problem with data ingestion. Please reach out to Roboto support."
        )



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import pathlib
import typing

T_co = typing.TypeVar("T_co", covariant=True)


class FutureLike(typing.Protocol[T_co]):
    """Protocol for future-like objects."""

    def result(self) -> T_co:
        """Wait for and return the result."""
        ...

    def done(self) -> bool:
        """Return True if the future is done."""
        ...


"""Callback which takes a number of bytes transferred to be periodically called."""
OnProgress: typing.TypeAlias = typing.Callable[[int], None]


@typing.runtime_checkable
class ObjectStore(typing.Protocol):
    @classmethod
    def create(cls, credential_provider: CredentialProvider, **kwargs) -> ObjectStore: ...

    def __enter__(self) -> ObjectStore: ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...

    def put(
        self, source: pathlib.Path, destination_uri: str, on_progress: typing.Optional[OnProgress] = None
    ) -> FutureLike[None]:
        """
        Uploads a local file to a specific cloud URI.

        Args:
            source: Local path to the file.
            destination_uri: Full URI (e.g., 's3://my-bucket/folder/data.csv')
            on_progress: Optional callback to be periodically called with the number of bytes uploaded.
        """
        ...

    def get(
        self, source_uri: str, destination: pathlib.Path, on_progress: typing.Optional[OnProgress] = None
    ) -> FutureLike[None]:
        """
        Downloads a file from a cloud URI to a local path.

        Parent directories of the destination path are created automatically if they don't exist.

        Args:
            source_uri: Full URI (e.g., 's3://my-bucket/folder/data.csv')
            destination: Local path where the file should be saved.
            on_progress: Optional callback to be periodically called with the number of bytes downloaded.
        """
        ...


class Credentials(typing.TypedDict):
    """
    This interface is driven by botocore.credentials.RefreshableCredentials
    """

    access_key: str
    secret_key: str
    token: str
    region: str
    expiry_time: typing.Optional[str]


CredentialProvider: typing.TypeAlias = typing.Callable[[], Credentials]



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .arrow_to_roboto import (
    generate_message_path_requests,
)
from .ingestion import (
    make_topic_filename_safe,
    upload_representation_file,
)
from .parquet_parser import ParquetParser
from .parquet_topic_reader import (
    ParquetTopicReader,
)

__all__ = (
    "generate_message_path_requests",
    "make_topic_filename_safe",
    "ParquetParser",
    "ParquetTopicReader",
    "upload_representation_file",
)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import importlib.metadata

try:
    from .autogen_version import AUTOGEN_VERSION

    __version__ = AUTOGEN_VERSION
except ImportError:
    __version__ = "0.0.0"


def roboto_version() -> str:
    try:
        return importlib.metadata.version("roboto")
    except importlib.metadata.PackageNotFoundError:
        return "version_not_found"


__all__ = ["__version__", "roboto_version"]



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import pathlib
import typing

from ...compat import import_optional_dependency
from ...http import RobotoClient
from ...logging import default_logger
from ...time import (
    Time,
    to_epoch_nanoseconds,
)
from .mcap_topic_reader import McapTopicReader
from .operations import (
    MessagePathRepresentationMapping,
)
from .parquet import ParquetTopicReader
from .record import (
    CanonicalDataType,
    MessagePathRecord,
)
from .topic_reader import Timestamp, TopicReader

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep

logger = default_logger()


def iterable_to_set(arg: collections.abc.Iterable[str]) -> set[str]:
    if isinstance(arg, set):
        return arg
    elif isinstance(arg, str):
        return {
            arg,
        }
    elif isinstance(arg, collections.abc.Iterable):
        return set(arg)
    else:
        raise TypeError("Input must be iterable.")


class TopicDataService:
    """Internal service for retrieving topic data.

    This service handles the low-level operations for accessing topic data that has been
    ingested by the Roboto platform. It manages downloads, filtering, and processing
    various data formats to provide efficient access to time-series robotics data.

    Note:
        This is not intended as a public API.
        To access topic data, prefer the ``get_data`` or ``get_data_as_df`` methods
        on :py:class:`~roboto.domain.topics.Topic`, :py:class:`~roboto.domain.topics.MessagePath`,
        or :py:class:`~roboto.domain.events.Event`.
    """

    DEFAULT_CACHE_DIR: typing.ClassVar[pathlib.Path] = pathlib.Path.home() / ".cache" / "roboto" / "topic-data"

    __cache_dir: pathlib.Path
    __roboto_client: RobotoClient

    def __init__(
        self,
        roboto_client: RobotoClient,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ):
        self.__roboto_client = roboto_client
        self.__cache_dir = pathlib.Path(cache_dir) if cache_dir is not None else TopicDataService.DEFAULT_CACHE_DIR

    def get_data(
        self,
        topic_id: str,
        message_paths_include: typing.Optional[collections.abc.Iterable[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Iterable[str]] = None,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir_override: typing.Union[str, pathlib.Path, None] = None,
    ) -> collections.abc.Generator[tuple[Timestamp, dict[str, typing.Any]], None, None]:
        """Retrieve data for a specific topic with optional filtering.

        Args:
            topic_id: Unique identifier of the topic to retrieve data for.
            message_paths_include: Dot notation paths to include in the results.
                If None, all paths are included.
            message_paths_exclude: Dot notation paths to exclude from the results.
                If None, no paths are excluded.
            start_time: Start time (inclusive) for temporal filtering.
            end_time: End time (exclusive) for temporal filtering.
            cache_dir_override: Override the default cache directory for downloads.

        Yields:
            Tuple of (timestamp, record) where timestamp is in nanoseconds since Unix epoch.
        """
        cache_dir = self.__ensure_cache_dir(cache_dir_override)

        message_path_repr_mappings = self.__get_message_path_mappings(topic_id)
        filtered_message_path_repr_mappings = self.__filter_message_path_mappings(
            message_path_repr_mappings,
            message_paths_include=message_paths_include,
            message_paths_exclude=message_paths_exclude,
        )

        start_time_ns = to_epoch_nanoseconds(start_time) if start_time is not None else None
        end_time_ns = to_epoch_nanoseconds(end_time) if end_time is not None else None

        if McapTopicReader.accepts(filtered_message_path_repr_mappings):
            reader: TopicReader = McapTopicReader(self.__roboto_client, cache_dir)
            yield from reader.get_data(
                filtered_message_path_repr_mappings,
                start_time=start_time_ns,
                end_time=end_time_ns,
            )
        elif ParquetTopicReader.accepts(filtered_message_path_repr_mappings):
            reader = ParquetTopicReader(self.__roboto_client, cache_dir=cache_dir)
            timestamp_mapping = self.__find_timestamp_message_path_mapping(message_path_repr_mappings)
            yield from reader.get_data(
                filtered_message_path_repr_mappings,
                start_time=start_time_ns,
                end_time=end_time_ns,
                timestamp_message_path_representation_mapping=timestamp_mapping,
            )
        else:
            raise NotImplementedError("No compatible reader found for this data. Please reach out to Roboto support.")

    def get_data_as_df(
        self,
        topic_id: str,
        message_paths_include: typing.Optional[collections.abc.Iterable[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Iterable[str]] = None,
        start_time: typing.Optional[Time] = None,
        end_time: typing.Optional[Time] = None,
        cache_dir_override: typing.Union[str, pathlib.Path, None] = None,
    ) -> pandas.DataFrame:
        """Retrieve data for a specific topic as a pandas DataFrame with optional filtering.

        Args:
            topic_id: Unique identifier of the topic to retrieve data for.
            message_paths_include: Dot notation paths to include in the results.
                If None, all paths are included.
            message_paths_exclude: Dot notation paths to exclude from the results.
                If None, no paths are excluded.
            start_time: Start time (inclusive) for temporal filtering.
            end_time: End time (exclusive) for temporal filtering.
            cache_dir_override: Override the default cache directory for downloads.

        Returns:
            pandas.DataFrame

            The index of the DataFrame (``df.index``) is a pandas.DateTimeIndex, labeling the timestamp of each row.
        """
        pd = import_optional_dependency("pandas", "analytics")

        cache_dir = self.__ensure_cache_dir(cache_dir_override)

        message_path_repr_mappings = self.__get_message_path_mappings(topic_id)
        filtered_message_path_repr_mappings = self.__filter_message_path_mappings(
            message_path_repr_mappings,
            message_paths_include=message_paths_include,
            message_paths_exclude=message_paths_exclude,
        )

        start_time_ns = to_epoch_nanoseconds(start_time) if start_time is not None else None
        end_time_ns = to_epoch_nanoseconds(end_time) if end_time is not None else None

        if McapTopicReader.accepts(filtered_message_path_repr_mappings):
            reader: TopicReader = McapTopicReader(self.__roboto_client, cache_dir)
            timestamps, df = reader.get_data_as_df(
                filtered_message_path_repr_mappings,
                start_time=start_time_ns,
                end_time=end_time_ns,
            )
        elif ParquetTopicReader.accepts(filtered_message_path_repr_mappings):
            reader = ParquetTopicReader(self.__roboto_client, cache_dir=cache_dir)
            timestamp_mapping = self.__find_timestamp_message_path_mapping(message_path_repr_mappings)
            timestamps, df = reader.get_data_as_df(
                filtered_message_path_repr_mappings,
                start_time=start_time_ns,
                end_time=end_time_ns,
                timestamp_message_path_representation_mapping=timestamp_mapping,
            )
        else:
            raise NotImplementedError("No compatible reader found for this data. Please reach out to Roboto support.")

        timestamps_as_datetimes = pd.to_datetime(timestamps, unit="ns", utc=True)
        df = df.set_index(timestamps_as_datetimes)
        # GM(2025-12-31)
        # When this dataframe is passed to Topic.create_from_df(), the index becomes a Parquet column.
        # Naming it avoids an even more cryptic, autogenerated MessagePath name like "__index__level_0__".
        # It's given a leading underscore to signify it is semi-private,
        # and also in attempt to avoid a name conflict with a column named "index".
        df.index.name = "_index"

        return df

    def __ensure_cache_dir(self, cache_dir_override: typing.Union[str, pathlib.Path, None] = None) -> pathlib.Path:
        cache_dir = pathlib.Path(cache_dir_override) if cache_dir_override is not None else self.__cache_dir
        if not cache_dir.exists():
            cache_dir.mkdir(parents=True)

        return cache_dir

    def __filter_message_paths(
        self,
        message_path_records: collections.abc.Iterable[MessagePathRecord],
        include_paths: typing.Optional[collections.abc.Iterable[str]],
        exclude_paths: typing.Optional[collections.abc.Iterable[str]],
    ) -> list[MessagePathRecord]:
        if not include_paths and not exclude_paths:
            return list(message_path_records)

        filtered = []
        include_paths_set = iterable_to_set(include_paths) if include_paths is not None else set()
        exclude_paths_set = iterable_to_set(exclude_paths) if exclude_paths is not None else set()

        for record in message_path_records:
            paths = set(record.parents())
            paths.add(record.source_path)
            paths.add(
                # May be different from source_path, support specifying either one
                record.message_path
            )

            if include_paths and paths.isdisjoint(include_paths_set):
                continue

            if exclude_paths and not paths.isdisjoint(exclude_paths_set):
                continue

            filtered.append(record)

        return filtered

    def __filter_message_path_mappings(
        self,
        message_path_repr_mappings: collections.abc.Iterable[MessagePathRepresentationMapping],
        message_paths_include: typing.Optional[collections.abc.Iterable[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Iterable[str]] = None,
    ) -> list[MessagePathRepresentationMapping]:
        # Exclude a MessagePathRepresentationMapping if no message paths remain after applying message path filters.
        message_path_repr_mappings = [
            message_path_repr_map.model_copy(
                update={
                    "message_paths": self.__filter_message_paths(
                        message_path_repr_map.message_paths,
                        include_paths=message_paths_include,
                        exclude_paths=message_paths_exclude,
                    )
                }
            )
            for message_path_repr_map in message_path_repr_mappings
        ]
        message_path_repr_mappings = [
            message_path_repr_map
            for message_path_repr_map in message_path_repr_mappings
            if message_path_repr_map.message_paths
        ]

        # WARN if no message paths are left after filtering
        if not any(
            len(message_path_repr_map.message_paths) > 0 for message_path_repr_map in message_path_repr_mappings
        ):
            logger.warning(
                "The request for topic data will not yield any results. "
                "Please check that the 'message_paths_include' and 'message_paths_exclude' parameters "
                "are not filtering out all available message paths for this topic."
            )

        return message_path_repr_mappings

    def __find_timestamp_message_path_mapping(
        self,
        message_path_repr_mappings: collections.abc.Iterable[MessagePathRepresentationMapping],
    ) -> typing.Optional[MessagePathRepresentationMapping]:
        for mapping in message_path_repr_mappings:
            for message_path in mapping.message_paths:
                if message_path.canonical_data_type != CanonicalDataType.Timestamp:
                    continue

                return MessagePathRepresentationMapping(
                    message_paths=[message_path], representation=mapping.representation
                )

        return None

    def __get_message_path_mappings(
        self,
        topic_id: str,
    ) -> list[MessagePathRepresentationMapping]:
        message_path_repr_mappings_response = self.__roboto_client.get(
            f"v1/topics/id/{topic_id}/message-path/representations"
        )
        return message_path_repr_mappings_response.to_record_list(MessagePathRepresentationMapping)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os
import pathlib
import re
import typing

import pydantic
import pydantic_settings

from .compat import StrEnum
from .sentinels import NotSet, NotSetType

ROBOTO_ENV_VAR_PREFIX = "ROBOTO_"

Timeout: typing.TypeAlias = typing.Optional[typing.Union[float, NotSetType]]
"""Timeout in seconds.

    NotSet: Use the client's default timeout
    None: No timeout (wait indefinitely)
    float: Explicit timeout in seconds
"""


def resolve_env_variables(value: str):
    """
    Given any input string, resolves any environment variables in the string against the current environment.
    """
    resolved_value = value

    # Left side of | is $MY_ENV_VAR style, right side is ${MY_ENV_VAR}
    #
    # Each side has the actual variable name inside an inner capture group (separate from $ or ${}) to make it easier
    # to pass to os.getenv
    pattern = re.compile(r"(\$(\w+))|(\$\{(\w+)})")
    matches = pattern.findall(resolved_value)

    for match in matches:
        left, left_word, right, right_word = match

        if left != "" and left_word != "":
            resolved_value = resolved_value.replace(left, os.getenv(left_word, ""))

        if right != "" and right_word != "":
            resolved_value = resolved_value.replace(right, os.getenv(right_word, ""))

    return resolved_value


class RobotoEnvKey(StrEnum):
    """Environment variable keys available within an action invocation runtime"""

    ActionInputsManifest = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_INPUTS_MANIFEST"
    ActionParametersFile = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_PARAMETERS_FILE"
    ActionRuntimeConfigDir = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_RUNTIME_CONFIG_DIR"
    ActionTimeout = f"{ROBOTO_ENV_VAR_PREFIX}ACTION_TIMEOUT"
    ApiKey = f"{ROBOTO_ENV_VAR_PREFIX}API_KEY"
    BearerToken = f"{ROBOTO_ENV_VAR_PREFIX}BEARER_TOKEN"
    """Interchangable with ApiKey"""
    ConfigFile = f"{ROBOTO_ENV_VAR_PREFIX}CONFIG_FILE"
    CacheDir = f"{ROBOTO_ENV_VAR_PREFIX}CACHE_DIR"
    DatasetMetadataChangesetFile = f"{ROBOTO_ENV_VAR_PREFIX}DATASET_METADATA_CHANGESET_FILE"
    DatasetId = f"{ROBOTO_ENV_VAR_PREFIX}DATASET_ID"
    DryRun = f"{ROBOTO_ENV_VAR_PREFIX}DRY_RUN"
    FileMetadataChangesetFile = f"{ROBOTO_ENV_VAR_PREFIX}FILE_METADATA_CHANGESET_FILE"
    InputDir = f"{ROBOTO_ENV_VAR_PREFIX}INPUT_DIR"
    InvocationId = f"{ROBOTO_ENV_VAR_PREFIX}INVOCATION_ID"
    LogLevel = f"{ROBOTO_ENV_VAR_PREFIX}LOG_LEVEL"
    OrgId = f"{ROBOTO_ENV_VAR_PREFIX}ORG_ID"
    OutputDir = f"{ROBOTO_ENV_VAR_PREFIX}OUTPUT_DIR"
    Profile = f"{ROBOTO_ENV_VAR_PREFIX}PROFILE"
    RobotoEnv = f"{ROBOTO_ENV_VAR_PREFIX}ENV"
    RobotoServiceUrl = f"{ROBOTO_ENV_VAR_PREFIX}SERVICE_URL"
    """Deprecated, use RobotoServiceEndpoint instead. Left here until 0.3.3 is released so we can migrate
    existing actions to use the new env var."""
    RobotoServiceEndpoint = f"{ROBOTO_ENV_VAR_PREFIX}SERVICE_ENDPOINT"

    @staticmethod
    def for_parameter(param_name: str) -> str:
        return f"{ROBOTO_ENV_VAR_PREFIX}PARAM_{param_name}"


_roboto_env_instance: typing.Optional["RobotoEnv"] = None


# You'll notice that the alias values here are duplicates of the RobotoEnvKey values. This is not ideal, but is
# necessary for type checking, because the value of alias needs to be a string literal and not a de-referenced variable.
# Even using f-strings will break the type checking.
class RobotoEnv(pydantic_settings.BaseSettings):
    """Enivronment within an action invocation runtime"""

    @classmethod
    def default(cls) -> "RobotoEnv":
        # We need to keep this module level vs. as a class variable, because otherwise pydantic will cast it to a
        # ModelPrivateAttr
        global _roboto_env_instance
        if _roboto_env_instance is None:
            _roboto_env_instance = cls()
        return _roboto_env_instance

    action_inputs_manifest_file: typing.Optional[pathlib.Path] = pydantic.Field(
        default=None, alias="ROBOTO_ACTION_INPUTS_MANIFEST"
    )

    action_parameters_file: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ACTION_PARAMETERS_FILE")

    action_runtime_config_dir: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_ACTION_RUNTIME_CONFIG_DIR"
    )

    action_timeout: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ACTION_TIMEOUT")

    api_key: typing.Optional[str] = pydantic.Field(
        default=None,
        validation_alias=pydantic.AliasChoices("ROBOTO_API_KEY", "ROBOTO_BEARER_TOKEN"),
    )

    cache_dir: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_CACHE_DIR")

    config_file: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_CONFIG_FILE")
    """
    An override for the location of a roboto config file. If not provided, the default ~/.roboto/config.json will be
    used (subject to RobotoConfig's implementation)
    """

    dataset_id: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_DATASET_ID")

    dataset_metadata_changeset_file: typing.Optional[str] = pydantic.Field(
        default=None,
        alias="ROBOTO_DATASET_METADATA_CHANGESET_FILE",
    )

    default_http_timeout: Timeout = pydantic.Field(default=NotSet, alias="ROBOTO_DEFAULT_HTTP_TIMEOUT")
    """
    Give up on Roboto Platform HTTP requests that take longer than this many seconds to complete.
    If the request is known to be idempotent, it will be automatically retried.
    Set to ``None`` to wait indefinitely.
    """

    dry_run: typing.Optional[bool] = pydantic.Field(default=False, alias="ROBOTO_DRY_RUN")
    """
    Flag that action developers should use to gate side effects during local invocation.

    When set to True, actions should skip operations that have side effects, such as:
    - Uploading files to Roboto datasets
    - Modifying metadata
    - Making external API calls that are not idempotent

    This enables safe local testing and development without affecting production resources.
    """

    file_metadata_changeset_file: typing.Optional[str] = pydantic.Field(
        default=None, alias="ROBOTO_FILE_METADATA_CHANGESET_FILE"
    )

    input_dir: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_INPUT_DIR")

    invocation_id: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_INVOCATION_ID")

    log_level: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_LOG_LEVEL")

    org_id: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ORG_ID")

    output_dir: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_OUTPUT_DIR")

    profile: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_PROFILE")
    """
    The profile name to use if getting RobotoConfig from a config file.
    """

    roboto_env: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_ENV")

    roboto_service_url: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_SERVICE_URL")
    """Deprecated, use roboto_service_endpoint instead. Left here until 0.3.3 is released so we can migrate
    existing actions to use the new env var."""

    roboto_service_endpoint: typing.Optional[str] = pydantic.Field(default=None, alias="ROBOTO_SERVICE_ENDPOINT")
    """A Roboto Service API endpoint to send requests to, typically https://api.roboto.ai"""

    def get_env_var(self, var_name: str, default_value: typing.Optional[str] = None) -> typing.Optional[str]:
        return os.getenv(var_name, default_value)



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
import typing

import tqdm


class ProgressMonitor(abc.ABC):
    """Progress Monitor for file upload"""

    @abc.abstractmethod
    def update(self, delta: int):
        raise NotImplementedError("update")

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError("close")

    @abc.abstractmethod
    def is_closed(self) -> bool:
        raise NotImplementedError("is_closed")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class NoopProgressMonitor(ProgressMonitor):
    """A Noop Progress Monitor"""

    def update(self, uploaded_bytes: int):
        pass

    def close(self):
        pass

    def is_closed(self) -> bool:
        return False


class TqdmProgressMonitor(ProgressMonitor):
    """A Tqdm Progress Monitor"""

    __tqdm: tqdm.tqdm
    __is_closed: bool

    def __init__(
        self,
        total: int,
        desc: str,
        position: int = 0,
        leave: bool = True,
        unit: typing.Optional[str] = "B",
    ):
        tqdm_args = {
            "total": total,
            "desc": desc,
            "bar_format": "{percentage:.1f}%|{bar:25} | {n_fmt}/{total_fmt} | {rate_fmt} | {elapsed} | {desc}",
            "position": position,
            "leave": leave,
        }

        if unit is not None:
            tqdm_args["unit"] = unit
            if unit == "B":
                tqdm_args["unit_scale"] = True
                tqdm_args["unit_divisor"] = 1024

        self.__tqdm = tqdm.tqdm(**tqdm_args)
        self.__is_closed = False

    def update(self, uploaded_bytes: int):
        self.__tqdm.update(uploaded_bytes)

    def close(self):
        self.__tqdm.close()
        self.__is_closed = True

    def is_closed(self) -> bool:
        return self.__is_closed



# Copyright (c) 2025 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import shutil
import sys

from ..command import RobotoCommand
from ..context import CLIContext


def clear(args, context: CLIContext, parser: argparse.ArgumentParser):
    cache_dir = context.roboto_config.get_cache_dir()

    if not cache_dir.exists():
        print(f"Cache directory does not exist: {cache_dir}", file=sys.stderr)
        return

    if args.dry_run:
        print(f"Would clear cache directory: {cache_dir}", file=sys.stderr)
        return

    shutil.rmtree(cache_dir)
    print(f"Cleared cache directory: {cache_dir}", file=sys.stderr)


def clear_setup_parser(parser):
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print location of cache directory that would be deleted, without performing the delete.",
    )


clear_command = RobotoCommand(
    name="clear",
    logic=clear,
    setup_parser=clear_setup_parser,
    command_kwargs={"help": "Clears the local Roboto cache directory."},
)
